# -*- coding: utf-8; test-case-name: bridgedb.test.test_https_request; -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft <isis@torproject.org> 0xA3ADB67A2CDB8B35
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2013-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

"""
.. py:module:: bridgedb.https.request
    :synopsis: Classes for parsing and storing information about requests for
               bridges which are sent to the HTTPS distributor.

bridgedb.https.request
======================

Classes for parsing and storing information about requests for bridges
which are sent to the HTTPS distributor.

.. inheritance-diagram:: HTTPSBridgeRequest

::

  bridgedb.https.request
   |
   |_ HTTPSBridgeRequest - A request for bridges which was received through
                           the HTTPS distributor.

..
"""

from __future__ import print_function
from __future__ import unicode_literals

import ipaddr
import logging
import re

from bridgedb import bridgerequest
from bridgedb import geo
from bridgedb.parse import addr


#: A regular expression for matching the Pluggable Transport methodname in
#: HTTP GET request parameters.
TRANSPORT_REGEXP = "[_a-zA-Z][_a-zA-Z0-9]*"
TRANSPORT_PATTERN = re.compile(TRANSPORT_REGEXP)

UNBLOCKED_REGEXP = "[a-zA-Z]{2}"
UNBLOCKED_PATTERN = re.compile(UNBLOCKED_REGEXP)


class HTTPSBridgeRequest(bridgerequest.BridgeRequestBase):
    """We received a request for bridges through the HTTPS distributor."""

    def __init__(self, addClientCountryCode=True):
        """Process a new bridge request received through the
        :class:`~bridgedb.https.distributor.HTTPSDistributor`.

        :param bool addClientCountryCode: If ``True``, then calling
            :meth:`withoutBlockInCountry` will attempt to add the client's own
            country code, geolocated from her IP, to the ``notBlockedIn``
            countries list.
        """
        super(HTTPSBridgeRequest, self).__init__()
        self.addClientCountryCode = addClientCountryCode

    def withIPversion(self, parameters):
        """Determine if the request **parameters** were for bridges with IPv6
        addresses or not.

        .. note:: If there is an ``ipv6=`` parameter with anything non-zero
            after it, then we assume the client wanted IPv6 bridges.

        :param parameters: The :api:`twisted.web.http.Request.args`.
        """
        if parameters.get("ipv6", False):
            logging.info("HTTPS request for bridges with IPv6 addresses.")
            self.withIPv6()

    def withoutBlockInCountry(self, request):
        """Determine which countries the bridges for this **request** should
        not be blocked in.

        .. note:: Currently, a **request** for unblocked bridges is recognized
            if it contains an HTTP GET parameter ``unblocked=`` whose value is
            a comma-separater list of two-letter country codes.  Any
            two-letter country code found in the
            :api:`request <twisted.web.http.Request>` ``unblocked=`` HTTP GET
            parameter will be added to the :data:`notBlockedIn` list.

        If :data:`addClientCountryCode` is ``True``, the the client's own
        geolocated country code will be added to the to the
        :data`notBlockedIn` list.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object containing the HTTP method, full
            URI, and any URL/POST arguments and headers present.
        """
        countryCodes = request.args.get("unblocked", list())

        for countryCode in countryCodes:
            try:
                country = UNBLOCKED_PATTERN.match(countryCode).group()
            except (TypeError, AttributeError):
                pass
            else:
                if country:
                    self.notBlockedIn.append(country.lower())
                    logging.info("HTTPS request for bridges not blocked in: %r"
                                 % country)

        if self.addClientCountryCode:
            # Look up the country code of the input IP, and request bridges
            # not blocked in that country.
            if addr.isIPAddress(self.client):
                country = geo.getCountryCode(ipaddr.IPAddress(self.client))
                if country:
                    self.notBlockedIn.append(country.lower())
                    logging.info(
                        ("HTTPS client's bridges also shouldn't be blocked "
                         "in their GeoIP country code: %s") % country)

    def withPluggableTransportType(self, parameters):
        """This request included a specific Pluggable Transport identifier.

        Add any Pluggable Transport methodname found in the HTTP GET
        **parameters** to the list of ``transports``. Currently, a request for
        a transport is recognized if the request contains the
        ``'transport='`` parameter.

        :param parameters: The :api:`twisted.web.http.Request.args`.
        """
        for methodname in parameters.get("transport", list()):
            try:
                transport = TRANSPORT_PATTERN.match(methodname).group()
            except (TypeError, AttributeError):
                pass
            else:
                if transport:
                    self.transports.append(transport)
                    logging.info("HTTPS request for transport type: %r"
                                 % transport)



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
# :copyright: (c) 2014-2015, Isis Lovecruft
#             (c) 2014-2015, The Tor Project, Inc.
# :license: 3-Clause BSD, see LICENSE for licensing information


from twisted.trial import unittest

from bridgedb import translations
from .https_helpers import DummyRequest


REALISH_HEADERS = {
    b'Accept-Encoding': [b'gzip, deflate'],
    b'User-Agent': [
        b'Mozilla/5.0 (X11; Linux x86_64; rv:28.0) Gecko/20100101 Firefox/28.0'],
    b'Accept': [
        b'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8'],
}

# Add this to the above REALISH_HEADERS to use it:
ACCEPT_LANGUAGE_HEADER = {
    b'Accept-Language': [b'de-de,en-gb;q=0.8,en;q=0.5,en-us;q=0.3'],
}


class TranslationsMiscTests(unittest.TestCase):
    """Tests for module-level code in ``bridgedb.translations`` module."""

    def test_getLocaleFromHTTPRequest_withLangParam(self):
        """This request uses a '?lang=ar' param, without an 'Accept-Language'
        header.

        The request result should be: ['ar', 'en', 'en-US'].
        """
        request = DummyRequest([b"bridges"])
        request.headers.update(REALISH_HEADERS)
        request.args.update({
            b'transport': [b'obfs3',],
            b'lang': [b'ar',],
        })

        parsed = translations.getLocaleFromHTTPRequest(request)
        self.assertEqual(parsed[0], 'ar')
        self.assertEqual(parsed[1], 'en')
        self.assertEqual(parsed[2], 'en_US')
        self.assertEqual(len(parsed), 3)

    def test_getLocaleFromHTTPRequest_withLangParam_AcceptLanguage(self):
        """This request uses a '?lang=ar' param, with an 'Accept-Language'
        header which includes: ['de-de', 'en-gb', 'en', 'en-us'].

        The request result should be: ['fa', 'de-de', 'en-gb', 'en', 'en-us'].
        """
        request = DummyRequest([b"options"])
        request.headers.update(ACCEPT_LANGUAGE_HEADER)
        request.args.update({b'lang': [b'fa']})

        parsed = translations.getLocaleFromHTTPRequest(request)
        self.assertEqual(parsed[0], 'fa')
        self.assertEqual(parsed[1], 'en')
        self.assertEqual(parsed[2], 'en_US')
        #self.assertEqual(parsed[3], 'en-gb')
        self.assertEqual(len(parsed), 3)

    def test_getLocaleFromPlusAddr(self):
        emailAddr = 'bridges@torproject.org'
        replyLocale = translations.getLocaleFromPlusAddr(emailAddr)
        self.assertEqual('en', replyLocale)

    def test_getLocaleFromPlusAddr_ar(self):
        emailAddr = 'bridges+ar@torproject.org'
        replyLocale = translations.getLocaleFromPlusAddr(emailAddr)
        self.assertEqual('ar', replyLocale)



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013, Isis Lovecruft
#             (c) 2007-2013, The Tor Project, Inc.
#             (c) 2007-2013, all entities within the AUTHORS file
# :license: 3-Clause BSD, see LICENSE for licensing information

"""Unittests for the :mod:`bridgedb.persistent` module.

These tests ensure that :meth:`bridgedb.persistent.State.save`,
:func:`bridgedb.persistent.load`, and :meth:`bridgedb.persistent.State.load`
are all functioning as expected.

This module should not import :mod:`sure`.
"""

import os

from copy import deepcopy
from io    import StringIO

from twisted.trial import unittest

from bridgedb import persistent


TEST_CONFIG_FILE = StringIO(unicode("""\
BRIDGE_FILES = ['bridge-descriptors', 'bridge-descriptors.new']
LOGFILE = 'bridgedb.log'"""))


class StateSaveAndLoadTests(unittest.TestCase):
    """Test save() and load() of :mod:`~bridgedb.persistent`."""

    timeout = 15

    def setUp(self):
        configuration = {}
        TEST_CONFIG_FILE.seek(0)
        compiled = compile(TEST_CONFIG_FILE.read(), '<string>', 'exec')
        exec compiled in configuration
        config = persistent.Conf(**configuration)

        self.config = config
        self.state  = persistent.State(**config.__dict__)
        self.state.config = config
        self.state.statefile = os.path.abspath('bridgedb.state')

    def loadedStateAssertions(self, loadedState):
        # For some reason, twisted.trial.unittest.TestCase in Python2.6
        # doesn't have an 'assertIsNotNone' attribute...
        self.assertTrue(loadedState is not None)
        self.assertIsInstance(loadedState, persistent.State)
        self.assertNotIdentical(self.state, loadedState)
        self.assertNotEqual(self.state, loadedState)
        # For some reason, twisted.trial.unittest.TestCase in Python2.6
        # doesn't have an 'assertItemsEqual' attribute...
        self.assertEqual(self.state.__dict__.keys().sort(),
                         loadedState.__dict__.keys().sort())

    def savedStateAssertions(self, savedStatefile=None):
        self.assertTrue(os.path.isfile(str(self.state.statefile)))
        if savedStatefile:
            self.assertTrue(os.path.isfile(str(savedStatefile)))

    def test_init_with_STATEFILE(self):
        config = self.config
        setattr(config, 'STATEFILE', '~/foo.state')
        state = persistent.State(**config.__dict__)
        self.loadedStateAssertions(state)
        statefile = state.statefile
        self.assertTrue(statefile.endswith('foo.state'))

    def test_init_without_config(self):
        state = persistent.State(None)
        self.loadedStateAssertions(state)

    def test_init_with_config(self):
        state = persistent.State(self.config)
        self.loadedStateAssertions(state)

    def test_get_statefile(self):
        statefile = self.state._get_statefile()
        self.assertIsInstance(statefile, basestring)

    def test_set_statefile(self):
        self.state._set_statefile('bar.state')
        statefile = self.state._get_statefile()
        self.assertIsInstance(statefile, basestring)

    def test_set_statefile_new_dir(self):
        config = self.config
        setattr(config, 'STATEFILE', 'statefiles/foo.state')
        state = persistent.State(**config.__dict__)
        self.loadedStateAssertions(state)
        statefile = state.statefile
        self.assertTrue(statefile.endswith('foo.state'))

    def test_del_statefile(self):
        self.state._set_statefile('baz.state')
        self.state._del_statefile()
        statefile = self.state._get_statefile()
        self.assertIsNone(statefile)

    def test_save(self):
        self.state.save()
        self.savedStateAssertions()

    def test_stateSaveTempfile(self):
        savefile = self.mktemp()
        self.state.statefile = savefile
        self.state.save(savefile)
        savedStatefile = str(self.state.statefile)

    def test_stateLoadTempfile(self):
        savefile = self.mktemp()
        self.state.statefile = savefile
        self.assertTrue(self.state.statefile.endswith(savefile))
        self.state.save(savefile)
        self.savedStateAssertions(savefile)
        loadedState = self.state.load(savefile)
        self.loadedStateAssertions(loadedState)

    def test_stateSaveAndLoad(self):
        self.state.save()
        loadedState = self.state.load()
        self.loadedStateAssertions(loadedState)

    def test_load(self):
        self.state.save()
        loadedState = persistent.load()
        self.loadedStateAssertions(loadedState)

    def test_load_with_state(self):
        loadedState = persistent.load(self.state)
        self.loadedStateAssertions(loadedState)

    def test_load_with_None(self):
        persistent._setState(None)
        self.assertRaises(persistent.MissingState,
                          persistent.load, None)

    def test_load_with_statefile(self):
        self.assertRaises(persistent.MissingState,
                          self.state.load, 'quux.state')

    def test_load_with_statefile_opened(self):
        fh = open('quux.state', 'w+')
        self.assertRaises(persistent.MissingState, self.state.load, fh)
        fh.close()

    def test_load_with_statefile_object(self):
        self.assertRaises(persistent.MissingState, self.state.load, object)

    def test_load_without_statefile(self):
        persistent._setState(None)
        self.state.statefile = None
        self.assertRaises(persistent.MissingState,
                          persistent.load)



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_https_server -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: please see included AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2013-2015, Isis Lovecruft
# :license: see LICENSE for licensing information

"""
.. py:module:: bridgedb.https.server
    :synopsis: Servers which interface with clients and distribute bridges
               over HTTP(S).

bridgedb.https.server
=====================

Servers which interface with clients and distribute bridges over HTTP(S).

.. inheritance-diagram:: TranslatedTemplateResource IndexResource OptionsResource HowtoResource CaptchaProtectedResource GimpCaptchaProtectedResource ReCaptchaProtectedResource BridgesResource
    :parts: 1
"""

import base64
import gettext
import logging
import random
import re
import time
import os

from functools import partial

from ipaddr import IPv4Address

import mako.exceptions
from mako.template import Template
from mako.lookup import TemplateLookup

from twisted.internet import reactor
from twisted.internet.error import CannotListenError
from twisted.web import resource
from twisted.web import static
from twisted.web.server import NOT_DONE_YET
from twisted.web.server import Site
from twisted.web.util import redirectTo

from bridgedb import captcha
from bridgedb import crypto
from bridgedb import strings
from bridgedb import translations
from bridgedb import txrecaptcha
from bridgedb.https.request import HTTPSBridgeRequest
from bridgedb.parse import headers
from bridgedb.parse.addr import isIPAddress
from bridgedb.qrcodes import generateQR
from bridgedb.safelog import logSafely
from bridgedb.schedule import Unscheduled
from bridgedb.schedule import ScheduledInterval
from bridgedb.util import replaceControlChars


TEMPLATE_DIR = os.path.join(os.path.dirname(__file__), 'templates')
rtl_langs = ('ar', 'he', 'fa', 'gu_IN', 'ku')

# Setting `filesystem_checks` to False is recommended for production servers,
# due to potential speed increases. This means that the atimes of the Mako
# template files aren't rechecked every time the template is requested
# (otherwise, if they are checked, and the atime is newer, the template is
# recompiled). `collection_size` sets the number of compiled templates which
# are cached before the least recently used ones are removed. See:
# http://docs.makotemplates.org/en/latest/usage.html#using-templatelookup
lookup = TemplateLookup(directories=[TEMPLATE_DIR],
                        output_encoding='utf-8',
                        filesystem_checks=False,
                        collection_size=500)
logging.debug("Set template root to %s" % TEMPLATE_DIR)

#: This server's public, fully-qualified domain name.
SERVER_PUBLIC_FQDN = None


def setFQDN(fqdn, https=True):
    """Set the global :data:`SERVER_PUBLIC FQDN` variable.

    :param str fqdn: The public, fully-qualified domain name of the HTTP
        server that will serve this resource.
    :param bool https: If ``True``, then ``'https://'`` will be prepended to
        the FQDN.  This is primarily used to create a
        ``Content-Security-Policy`` header that will only allow resources to
        be sourced via HTTPS, otherwise, if ``False``, it allow resources to
        be sourced via any transport protocol.
    """
    if https:
        fqdn = 'https://' + fqdn

    logging.info("Setting HTTP server public FQDN to %r" % fqdn)

    global SERVER_PUBLIC_FQDN
    SERVER_PUBLIC_FQDN = fqdn

def getFQDN():
    """Get the setting for the HTTP server's public FQDN from the global
    :data:`SERVER_PUBLIC_FQDN variable.

    :rtype: str or None
    """
    return SERVER_PUBLIC_FQDN

def getClientIP(request, useForwardedHeader=False):
    """Get the client's IP address from the ``'X-Forwarded-For:'``
    header, or from the :api:`request <twisted.web.server.Request>`.

    :type request: :api:`twisted.web.http.Request`
    :param request: A ``Request`` for a :api:`twisted.web.resource.Resource`.
    :param bool useForwardedHeader: If ``True``, attempt to get the client's
        IP address from the ``'X-Forwarded-For:'`` header.
    :rtype: ``None`` or :any:`str`
    :returns: The client's IP address, if it was obtainable.
    """
    ip = None

    if useForwardedHeader:
        header = request.getHeader("X-Forwarded-For")
        if header:
            ip = header.split(",")[-1].strip()
            if not isIPAddress(ip):
                logging.warn("Got weird X-Forwarded-For value %r" % header)
                ip = None
    else:
        ip = request.getClientIP()

    return ip

def replaceErrorPage(request, error, template_name=None, html=True):
    """Create a general error page for displaying in place of tracebacks.

    Log the error to BridgeDB's logger, and then display a very plain "Sorry!
    Something went wrong!" page to the client.

    :type request: :api:`twisted.web.http.Request`
    :param request: A ``Request`` object containing the HTTP method, full
        URI, and any URL/POST arguments and headers present.
    :type error: :exc:`Exception`
    :param error: Any exeption which has occurred while attempting to retrieve
        a template, render a page, or retrieve a resource.
    :param str template_name: A string describing which template/page/resource
        was being used when the exception occurred, i.e. ``'index.html'``.
    :param bool html: If ``True``, return one of two HTML error pages.  First,
        we attempt to render a fancier error page.  If that rendering failed,
        or if **html** is ``False``, then we return a very simple HTML page
        (without CSS, Javascript, images, etc.)  which simply says
        ``"Sorry! Something went wrong with your request."``
    :rtype: str
    :returns: A string containing some content to serve to the client (rather
        than serving a Twisted traceback).
    """
    logging.error("Error while attempting to render %s: %s"
                  % (template_name or 'template',
                     mako.exceptions.text_error_template().render()))

    # TRANSLATORS: Please DO NOT translate the following words and/or phrases in
    # any string (regardless of capitalization and/or punctuation):
    #
    # "BridgeDB"
    # "pluggable transport"
    # "pluggable transports"
    # "obfs2"
    # "obfs3"
    # "scramblesuit"
    # "fteproxy"
    # "Tor"
    # "Tor Browser"
    #
    errorMessage = _("Sorry! Something went wrong with your request.")

    if not html:
        return bytes(errorMessage)

    try:
        rendered = resource500.render(request)
    except Exception as err:
        logging.exception(err)
        rendered = bytes(errorMessage)

    return rendered


class CSPResource(resource.Resource):
    """A resource which adds a ``'Content-Security-Policy:'`` header.

    :vartype reportViolations: bool
    :var reportViolations: Use the Content Security Policy in `report-only`_
        mode, causing CSP violations to be reported back to the server (at
        :attr:`reportURI`, where the details of the violation will be logged).
        (default: ``False``)
    :vartype reportURI: str
    :var reportURI: If :attr:`reportViolations` is ``True``, Content Security
        Policy violations will be sent as JSON-encoded POST request to this
        URI.  (default: ``'csp-violation'``)

    .. _report-only:
        https://w3c.github.io/webappsec/specs/content-security-policy/#content-security-policy-report-only-header-field
    """
    reportViolations = False
    reportURI = 'csp-violation'

    def __init__(self, includeSelf=False, enabled=True, reportViolations=False,
                 useForwardedHeader=False):
        """Create a new :api:`twisted.web.resource.Resource` which adds a
        ``'Content-Security-Policy:'`` header.

        If enabled, the default Content Security Policy is::

            default-src 'none' ;
            base-uri FQDN ;
            script-src FQDN ;
            style-src FQDN ;
            img-src FQDN data: ;
            font-src FQDN ;

        where ``FQDN`` the value returned from the :func:`getFQDN` function
        (which uses the ``SERVER_PUBLIC_FQDN`` config file option).

        If the **includeSelf** parameter is enabled, then ``"'self'"``
        (literally, a string containing the word ``self``, surrounded by
        single-quotes) will be appended to the ``FQDN``.

        :param str fqdn: The public, fully-qualified domain name
            of the HTTP server that will serve this resource.
        :param bool includeSelf: Append ``'self'`` after the **fqdn** in the
            Content Security Policy.
        :param bool enabled: If ``False``, all Content Security Policy
            headers, including those used in report-only mode, will not be
            sent.  If ``True``, Content Security Policy headers (regardless of
            whether report-only mode is dis-/en-abled) will be sent.
            (default: ``True``)
        :param bool reportViolations: Use the Content Security Policy in
            report-only mode, causing CSP violations to be reported back to
            the server (at :attr:`reportURI`, where the details of the
            violation will be logged).  (default: ``False``)
        :param bool useForwardedHeader: If ``True``, then we will attempt to
            obtain the client's IP address from the ``X-Forwarded-For`` HTTP
            header.  This *only* has an effect if **reportViolations** is also
            set to ``True`` — the client's IP address is logged along with any
            CSP violation reports which the client sent via HTTP POST requests
            to our :attr:`reportURI`.  (default: ``False``)
        """
        resource.Resource.__init__(self)

        self.fqdn = getFQDN()
        self.enabled = enabled
        self.useForwardedHeader = useForwardedHeader
        self.csp = ("default-src 'none'; "
                    "base-uri {0}; "
                    "script-src {0}; "
                    "style-src {0}; "
                    "img-src {0} data:; "
                    "font-src {0}; ")

        if includeSelf:
            self.fqdn = " ".join([self.fqdn, "'self'"])

        if reportViolations:
            self.reportViolations = reportViolations

    def setCSPHeader(self, request):
        """Set the CSP header for a **request**.

        If this :class:`CSPResource` is :attr:`enabled`, then use
        :api:`twisted.web.http.Request.setHeader` to send an HTTP
        ``'Content-Security-Policy:'`` header for any response made to the
        **request** (or a ``'Content-Security-Policy-Report-Only:'`` header,
        if :attr:`reportViolations` is enabled).

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for :attr:`reportViolationURI`.
        """
        self.fqdn = self.fqdn or getFQDN()  # Update the FQDN if it changed.

        if self.enabled and self.fqdn:
            if not self.reportViolations:
                request.setHeader("Content-Security-Policy",
                                  self.csp.format(self.fqdn))
            else:
                logging.debug("Sending report-only CSP header...")
                request.setHeader("Content-Security-Policy-Report-Only",
                                  self.csp.format(self.fqdn) +
                                  "report-uri /%s" % self.reportURI)

    def render_POST(self, request):
        """If we're in debug mode, log a Content Security Policy violation.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for :attr:`reportViolationURI`.
        """
        try:
            client = getClientIP(request, self.useForwardedHeader)
            report = request.content.read(2048)

            logging.warning("Content-Security-Policy violation report from %s: %r"
                            % (client or "UNKNOWN CLIENT", report))
        except Exception as err:
            logging.error("Error while attempting to log CSP report: %s" % err)

        # Redirect back to the original resource after the report was logged:
        return redirectTo(request.uri, request)


class ErrorResource(CSPResource):
    """A resource which explains that BridgeDB is undergoing maintenance, or
    that some other (unexpected) error has occured.
    """
    isLeaf = True

    def __init__(self, template=None, code=200):
        """Create a :api:`twisted.web.resource.Resource` for an error page."""
        CSPResource.__init__(self)
        self.template = template
        self.code = code

    def render_GET(self, request):
        self.setCSPHeader(request)
        request.setHeader("Content-Type", "text/html; charset=utf-8")
        request.setResponseCode(self.code)

        try:
            template = lookup.get_template(self.template)
            rendered = template.render()
        except Exception as err:
            rendered = replaceErrorPage(request, err, html=False)

        return rendered

    render_POST = render_GET

resource404 = ErrorResource('error-404.html', code=404)
resource500 = ErrorResource('error-500.html', code=500)
maintenance = ErrorResource('error-503.html', code=503)


class CustomErrorHandlingResource(resource.Resource):
    """A :api:`twisted.web.resource.Resource` which wraps the
    :api:`twisted.web.resource.Resource.getChild` method in order to use
    custom error handling pages.
    """
    def getChild(self, path, request):
        logging.debug("[404] %s" % request.uri)
        return resource404


class TranslatedTemplateResource(CustomErrorHandlingResource, CSPResource):
    """A generalised resource which uses gettext translations and Mako
    templates.
    """
    isLeaf = True

    def __init__(self, template=None):
        """Create a new :api:`Resource <twisted.web.resource.Resource>` for a
        Mako-templated webpage.
        """
        gettext.install("bridgedb", unicode=True)
        CSPResource.__init__(self)
        self.template = template

    def render_GET(self, request):
        self.setCSPHeader(request)
        rtl = False
        try:
            langs = translations.getLocaleFromHTTPRequest(request)
            rtl = translations.usingRTLLang(langs)
            template = lookup.get_template(self.template)
            rendered = template.render(strings, rtl=rtl, lang=langs[0])
        except Exception as err:  # pragma: no cover
            rendered = replaceErrorPage(request, err)
        request.setHeader("Content-Type", "text/html; charset=utf-8")
        return rendered

    render_POST = render_GET


class IndexResource(TranslatedTemplateResource):
    """The parent resource of all other documents hosted by the webserver."""

    def __init__(self):
        """Create a :api:`twisted.web.resource.Resource` for the index page."""
        TranslatedTemplateResource.__init__(self, 'index.html')


class OptionsResource(TranslatedTemplateResource):
    """A resource with additional options which a client may use to specify the
    which bridge types should be returned by :class:`BridgesResource`.
    """
    def __init__(self):
        """Create a :api:`twisted.web.resource.Resource` for the options page."""
        TranslatedTemplateResource.__init__(self, 'options.html')


class HowtoResource(TranslatedTemplateResource):
    """A resource which explains how to use bridges."""

    def __init__(self):
        """Create a :api:`twisted.web.resource.Resource` for the HowTo page."""
        TranslatedTemplateResource.__init__(self, 'howto.html')


class CaptchaProtectedResource(CustomErrorHandlingResource, CSPResource):
    """A general resource protected by some form of CAPTCHA."""

    isLeaf = True

    def __init__(self, publicKey=None, secretKey=None,
                 useForwardedHeader=False, protectedResource=None):
        CSPResource.__init__(self)
        self.publicKey = publicKey
        self.secretKey = secretKey
        self.useForwardedHeader = useForwardedHeader
        self.resource = protectedResource

    def getClientIP(self, request):
        """Get the client's IP address from the ``'X-Forwarded-For:'``
        header, or from the :api:`request <twisted.web.server.Request>`.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` for a
            :api:`twisted.web.resource.Resource`.
        :rtype: ``None`` or :any:`str`
        :returns: The client's IP address, if it was obtainable.
        """
        return getClientIP(request, self.useForwardedHeader)

    def getCaptchaImage(self, request=None):
        """Get a CAPTCHA image.

        :rtype: tuple
        :returns: A 2-tuple of ``(image, challenge)``, where ``image`` is a
            binary, JPEG-encoded image, and ``challenge`` is a unique
            string. If unable to retrieve a CAPTCHA, returns a tuple
            containing two empty strings.
        """
        return ('', '')

    def extractClientSolution(self, request):
        """Extract the client's CAPTCHA solution from a POST request.

        This is used after receiving a POST request from a client (which
        should contain their solution to the CAPTCHA), to extract the solution
        and challenge strings.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for 'bridges.html'.
        :returns: A redirect for a request for a new CAPTCHA if there was a
            problem. Otherwise, returns a 2-tuple of strings, the first is the
            client's CAPTCHA solution from the text input area, and the second
            is the challenge string.
        """
        try:
            challenge = request.args['captcha_challenge_field'][0]
            response = request.args['captcha_response_field'][0]
        except Exception:  # pragma: no cover
            return redirectTo(request.URLPath(), request)
        return (challenge, response)

    def checkSolution(self, request):
        """Override this method to check a client's CAPTCHA solution.

        :rtype: bool
        :returns: ``True`` if the client correctly solved the CAPTCHA;
            ``False`` otherwise.
        """
        return False

    def render_GET(self, request):
        """Retrieve a ReCaptcha from the API server and serve it to the client.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for a page which should be
            protected by a CAPTCHA.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
            for the client to solve.
        """
        self.setCSPHeader(request)

        rtl = False
        image, challenge = self.getCaptchaImage(request)

        try:
            langs = translations.getLocaleFromHTTPRequest(request)
            rtl = translations.usingRTLLang(langs)
            # TODO: this does not work for versions of IE < 8.0
            imgstr = 'data:image/jpeg;base64,%s' % base64.b64encode(image)
            template = lookup.get_template('captcha.html')
            rendered = template.render(strings,
                                       rtl=rtl,
                                       lang=langs[0],
                                       imgstr=imgstr,
                                       challenge_field=challenge)
        except Exception as err:
            rendered = replaceErrorPage(request, err, 'captcha.html')

        request.setHeader("Content-Type", "text/html; charset=utf-8")
        return rendered

    def render_POST(self, request):
        """Process a client's CAPTCHA solution.

        If the client's CAPTCHA solution is valid (according to
        :meth:`checkSolution`), process and serve their original
        request. Otherwise, redirect them back to a new CAPTCHA page.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
            for the client to solve.
        """
        self.setCSPHeader(request)
        request.setHeader("Content-Type", "text/html; charset=utf-8")

        if self.checkSolution(request) is True:
            try:
                rendered = self.resource.render(request)
            except Exception as err:
                rendered = replaceErrorPage(request, err)
            return rendered

        logging.debug("Client failed a CAPTCHA; returning redirect to %s"
                      % request.uri)
        return redirectTo(request.uri, request)


class GimpCaptchaProtectedResource(CaptchaProtectedResource):
    """A web resource which uses a local cache of CAPTCHAs, generated with
    gimp-captcha_, to protect another resource.

    .. _gimp-captcha: https://github.com/isislovecruft/gimp-captcha
    """

    def __init__(self, hmacKey=None, captchaDir='', **kwargs):
        """Protect a resource via this one, using a local CAPTCHA cache.

        :param str secretkey: A PKCS#1 OAEP-padded, private RSA key, used for
            verifying the client's solution to the CAPTCHA. See
            :func:`bridgedb.crypto.getRSAKey` and the
            ``GIMP_CAPTCHA_RSA_KEYFILE`` config setting.
        :param str publickey: A PKCS#1 OAEP-padded, public RSA key, used for
            creating the ``captcha_challenge_field`` string to give to a
            client.
        :param bytes hmacKey: The master HMAC key, used for validating CAPTCHA
            challenge strings in :meth:`captcha.GimpCaptcha.check`. The file
            where this key is stored can be set via the
            ``GIMP_CAPTCHA_HMAC_KEYFILE`` option in the config file.
        :param str captchaDir: The directory where the cached CAPTCHA images
            are stored. See the ``GIMP_CAPTCHA_DIR`` config setting.
        :param bool useForwardedHeader: If ``True``, obtain the client's IP
            address from the ``X-Forwarded-For`` HTTP header.
        :type protectedResource: :api:`twisted.web.resource.Resource`
        :param protectedResource: The resource to serve if the client
            successfully passes the CAPTCHA challenge.
        :param str serverPublicFQDN: The public, fully-qualified domain name
            of the HTTP server that will serve this resource.
        """
        CaptchaProtectedResource.__init__(self, **kwargs)
        self.hmacKey = hmacKey
        self.captchaDir = captchaDir

    def checkSolution(self, request):
        """Process a solved CAPTCHA via :meth:`bridgedb.captcha.GimpCaptcha.check`.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :rtupe: bool
        :returns: True, if the CAPTCHA solution was valid; False otherwise.
        """
        valid = False
        challenge, solution = self.extractClientSolution(request)
        clientIP = self.getClientIP(request)
        clientHMACKey = crypto.getHMAC(self.hmacKey, clientIP)

        try:
            valid = captcha.GimpCaptcha.check(challenge, solution,
                                              self.secretKey, clientHMACKey)
        except captcha.CaptchaExpired as error:
            logging.warn(error)
            valid = False

        logging.debug("%sorrect captcha from %r: %r."
                      % ("C" if valid else "Inc", clientIP, solution))
        return valid

    def getCaptchaImage(self, request):
        """Get a random CAPTCHA image from our **captchaDir**.

        Creates a :class:`~bridgedb.captcha.GimpCaptcha`, and calls its
        :meth:`~bridgedb.captcha.GimpCaptcha.get` method to return a random
        CAPTCHA and challenge string.

        :type request: :api:`twisted.web.http.Request`
        :param request: A client's initial request for some other resource
            which is protected by this one (i.e. protected by a CAPTCHA).
        :returns: A 2-tuple of ``(image, challenge)``, where::
            - ``image`` is a string holding a binary, JPEG-encoded image.
            - ``challenge`` is a unique string associated with the request.
        """
        # Create a new HMAC key, specific to requests from this client:
        clientIP = self.getClientIP(request)
        clientHMACKey = crypto.getHMAC(self.hmacKey, clientIP)
        capt = captcha.GimpCaptcha(self.publicKey, self.secretKey,
                                   clientHMACKey, self.captchaDir)
        try:
            capt.get()
        except captcha.GimpCaptchaError as error:
            logging.error(error)
        except Exception as error:  # pragma: no cover
            logging.error("Unhandled error while retrieving Gimp captcha!")
            logging.exception(error)

        return (capt.image, capt.challenge)

    def render_GET(self, request):
        """Get a random CAPTCHA from our local cache directory and serve it to
        the client.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for a page which should be
            protected by a CAPTCHA.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
           for the client to solve.
        """
        return CaptchaProtectedResource.render_GET(self, request)

    def render_POST(self, request):
        """Process a client's CAPTCHA solution.

        If the client's CAPTCHA solution is valid (according to
        :meth:`checkSolution`), process and serve their original
        request. Otherwise, redirect them back to a new CAPTCHA page.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
            for the client to solve.
        """
        return CaptchaProtectedResource.render_POST(self, request)


class ReCaptchaProtectedResource(CaptchaProtectedResource):
    """A web resource which uses the reCaptcha_ service.

    .. _reCaptcha: http://www.google.com/recaptcha
    """

    def __init__(self, remoteIP=None, **kwargs):
        CaptchaProtectedResource.__init__(self, **kwargs)
        self.remoteIP = remoteIP

    def _renderDeferred(self, checkedRequest):
        """Render this resource asynchronously.

        :type checkedRequest: tuple
        :param checkedRequest: A tuple of ``(bool, request)``, as returned
            from :meth:`checkSolution`.
        """
        try:
            valid, request = checkedRequest
        except Exception as err:
            logging.error("Error in _renderDeferred(): %s" % err)
            return

        logging.debug("Attemping to render %svalid request %r"
                      % ('' if valid else 'in', request))
        if valid is True:
            try:
                rendered = self.resource.render(request)
            except Exception as err:  # pragma: no cover
                rendered = replaceErrorPage(request, err)
        else:
            logging.info("Client failed a CAPTCHA; redirecting to %s"
                         % request.uri)
            rendered = redirectTo(request.uri, request)

        try:
            request.write(rendered)
            request.finish()
        except Exception as err:  # pragma: no cover
            logging.exception(err)

        return request

    def getCaptchaImage(self, request):
        """Get a CAPTCHA image from the remote reCaptcha server.

        :type request: :api:`twisted.web.http.Request`
        :param request: A client's initial request for some other resource
            which is protected by this one (i.e. protected by a CAPTCHA).
        :returns: A 2-tuple of ``(image, challenge)``, where::
            - ``image`` is a string holding a binary, JPEG-encoded image.
            - ``challenge`` is a unique string associated with the request.
        """
        capt = captcha.ReCaptcha(self.publicKey, self.secretKey)

        try:
            capt.get()
        except Exception as error:
            logging.fatal("Connection to Recaptcha server failed: %s" % error)

        if capt.image is None:
            logging.warn("No CAPTCHA image received from ReCaptcha server!")

        return (capt.image, capt.challenge)

    def getRemoteIP(self):
        """Mask the client's real IP address with a faked one.

        The fake client IP address is sent to the reCaptcha server, and it is
        either the public IP address of bridges.torproject.org (if the config
        option ``RECAPTCHA_REMOTE_IP`` is configured), or a random IP.

        :rtype: str
        :returns: A fake IP address to report to the reCaptcha API server.
        """
        if self.remoteIP:
            remoteIP = self.remoteIP
        else:
            # generate a random IP for the captcha submission
            remoteIP = IPv4Address(random.randint(0, 2**32-1)).compressed

        return remoteIP

    def checkSolution(self, request):
        """Process a solved CAPTCHA by sending it to the ReCaptcha server.

        The client's IP address is not sent to the ReCaptcha server; instead,
        a completely random IP is generated and sent instead.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :rtupe: :api:`twisted.internet.defer.Deferred`
        :returns: A deferred which will callback with a tuple in the following
            form:
                (:type:`bool`, :api:`twisted.web.server.Request`)
            If the CAPTCHA solution was valid, a tuple will contain::
                (True, Request)
            Otherwise, it will contain::
                (False, Request)
        """
        challenge, response = self.extractClientSolution(request)
        clientIP = self.getClientIP(request)
        remoteIP = self.getRemoteIP()

        logging.debug("Captcha from %r. Parameters: %r"
                      % (clientIP, request.args))

        def checkResponse(solution, request):
            """Check the :class:`txrecaptcha.RecaptchaResponse`.

            :type solution: :class:`txrecaptcha.RecaptchaResponse`.
            :param solution: The client's CAPTCHA solution, after it has been
                submitted to the reCaptcha API server.
            """
            # This valid CAPTCHA result from this function cannot be reliably
            # unittested, because it's callbacked to from the deferred
            # returned by ``txrecaptcha.submit``, the latter of which would
            # require networking (as well as automated CAPTCHA
            # breaking). Hence, the 'no cover' pragma.
            if solution.is_valid:  # pragma: no cover
                logging.info("Valid CAPTCHA solution from %r." % clientIP)
                return (True, request)
            else:
                logging.info("Invalid CAPTCHA solution from %r: %r"
                             % (clientIP, solution.error_code))
                return (False, request)

        d = txrecaptcha.submit(challenge, response, self.secretKey,
                               remoteIP).addCallback(checkResponse, request)
        return d

    def render_GET(self, request):
        """Retrieve a ReCaptcha from the API server and serve it to the client.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for 'bridges.html'.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
            for the client to solve.
        """
        return CaptchaProtectedResource.render_GET(self, request)

    def render_POST(self, request):
        """Process a client's CAPTCHA solution.

        If the client's CAPTCHA solution is valid (according to
        :meth:`checkSolution`), process and serve their original
        request. Otherwise, redirect them back to a new CAPTCHA page.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :returns: :api:`twisted.web.server.NOT_DONE_YET`, in order to handle
            the ``Deferred`` returned from :meth:`checkSolution`. Eventually,
            when the ``Deferred`` request is done being processed,
            :meth:`_renderDeferred` will handle rendering and displaying the
            HTML to the client.
        """
        self.setCSPHeader(request)
        d = self.checkSolution(request)
        d.addCallback(self._renderDeferred)
        return NOT_DONE_YET


class BridgesResource(CustomErrorHandlingResource, CSPResource):
    """This resource displays bridge lines in response to a request."""

    isLeaf = True

    def __init__(self, distributor, schedule, N=1, useForwardedHeader=False,
                 includeFingerprints=True):
        """Create a new resource for displaying bridges to a client.

        :type distributor: :class:`HTTPSDistributor`
        :param distributor: The mechanism to retrieve bridges for this
            distributor.
        :type schedule: :class:`~bridgedb.schedule.ScheduledInterval`
        :param schedule: The time period used to tweak the bridge selection
            procedure.
        :param int N: The number of bridges to hand out per query.
        :param bool useForwardedHeader: Whether or not we should use the the
            X-Forwarded-For header instead of the source IP address.
        :param bool includeFingerprints: Do we include the bridge's
            fingerprint in the response?
        """
        gettext.install("bridgedb", unicode=True)
        CSPResource.__init__(self)
        self.distributor = distributor
        self.schedule = schedule
        self.nBridgesToGive = N
        self.useForwardedHeader = useForwardedHeader
        self.includeFingerprints = includeFingerprints

    def render(self, request):
        """Render a response for a client HTTP request.

        Presently, this method merely wraps :meth:`getBridgeRequestAnswer` to
        catch any unhandled exceptions which occur (otherwise the server will
        display the traceback to the client). If an unhandled exception *does*
        occur, the client will be served the default "No bridges currently
        available" HTML response page.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object containing the HTTP method, full
            URI, and any URL/POST arguments and headers present.
        :rtype: str
        :returns: A plaintext or HTML response to serve.
        """
        self.setCSPHeader(request)

        try:
            response = self.getBridgeRequestAnswer(request)
        except Exception as err:
            logging.exception(err)
            response = self.renderAnswer(request)

        return response

    def getClientIP(self, request):
        """Get the client's IP address from the ``'X-Forwarded-For:'``
        header, or from the :api:`request <twisted.web.server.Request>`.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for a
            :api:`twisted.web.resource.Resource`.
        :rtype: ``None`` or :any:`str`
        :returns: The client's IP address, if it was obtainable.
        """
        return getClientIP(request, self.useForwardedHeader)

    def getBridgeRequestAnswer(self, request):
        """Respond to a client HTTP request for bridges.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object containing the HTTP method, full
            URI, and any URL/POST arguments and headers present.
        :rtype: str
        :returns: A plaintext or HTML response to serve.
        """
        bridgeLines = None
        interval = self.schedule.intervalStart(time.time())
        ip = self.getClientIP(request)

        logging.info("Replying to web request from %s. Parameters were %r"
                     % (ip, request.args))

        if ip:
            bridgeRequest = HTTPSBridgeRequest()
            bridgeRequest.client = ip
            bridgeRequest.isValid(True)
            bridgeRequest.withIPversion(request.args)
            bridgeRequest.withPluggableTransportType(request.args)
            bridgeRequest.withoutBlockInCountry(request)
            bridgeRequest.generateFilters()

            bridges = self.distributor.getBridges(bridgeRequest, interval)
            bridgeLines = [replaceControlChars(bridge.getBridgeLine(
                bridgeRequest, self.includeFingerprints)) for bridge in bridges]

        return self.renderAnswer(request, bridgeLines)

    def getResponseFormat(self, request):
        """Determine the requested format for the response.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object containing the HTTP method, full
            URI, and any URL/POST arguments and headers present.
        :rtype: ``None`` or str
        :returns: The argument of the first occurence of the ``format=`` HTTP
            GET parameter, if any were present. (The only one which currently
            has any effect is ``format=plain``, see note in
            :meth:`renderAnswer`.)  Otherwise, returns ``None``.
        """
        format = request.args.get("format", None)
        if format and len(format):
            format = format[0]  # Choose the first arg
        return format

    def renderAnswer(self, request, bridgeLines=None):
        """Generate a response for a client which includes **bridgesLines**.

        .. note: The generated response can be plain or HTML. A plain response
            looks like::

                voltron 1.2.3.4:1234 ABCDEF01234567890ABCDEF01234567890ABCDEF
                voltron 5.5.5.5:5555 0123456789ABCDEF0123456789ABCDEF01234567

            That is, there is no HTML, what you see is what you get, and what
            you get is suitable for pasting directly into Tor Launcher (or
            into a torrc, if you prepend ``"Bridge "`` to each line). The
            plain format can be requested from BridgeDB's web service by
            adding an ``&format=plain`` HTTP GET parameter to the URL. Also
            note that you won't get a QRCode, usage instructions, error
            messages, or any other fanciness if you use the plain format.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object containing the HTTP method, full
            URI, and any URL/POST arguments and headers present.
        :type bridgeLines: list or None
        :param bridgeLines: A list of strings used to configure a Tor client
            to use a bridge. If ``None``, then the returned page will instead
            explain that there were no bridges of the type they requested,
            with instructions on how to proceed.
        :rtype: str
        :returns: A plaintext or HTML response to serve.
        """
        rtl = False
        format = self.getResponseFormat(request)

        if format == 'plain':
            request.setHeader("Content-Type", "text/plain")
            try:
                rendered = bytes('\n'.join(bridgeLines))
            except Exception as err:
                rendered = replaceErrorPage(request, err, html=False)
        else:
            request.setHeader("Content-Type", "text/html; charset=utf-8")
            qrcode = None
            qrjpeg = generateQR(bridgeLines)

            if qrjpeg:
                qrcode = 'data:image/jpeg;base64,%s' % base64.b64encode(qrjpeg)
            try:
                langs = translations.getLocaleFromHTTPRequest(request)
                rtl = translations.usingRTLLang(langs)
                template = lookup.get_template('bridges.html')
                rendered = template.render(strings,
                                           rtl=rtl,
                                           lang=langs[0],
                                           answer=bridgeLines,
                                           qrcode=qrcode)
            except Exception as err:
                rendered = replaceErrorPage(request, err)

        return rendered


def addWebServer(config, distributor):
    """Set up a web server for HTTP(S)-based bridge distribution.

    :type config: :class:`bridgedb.persistent.Conf`
    :param config: A configuration object from
         :mod:`bridgedb.Main`. Currently, we use these options::
             HTTP_UNENCRYPTED_PORT
             HTTP_UNENCRYPTED_BIND_IP
             HTTP_USE_IP_FROM_FORWARDED_HEADER
             HTTPS_N_BRIDGES_PER_ANSWER
             HTTPS_INCLUDE_FINGERPRINTS
             HTTPS_KEY_FILE
             HTTPS_CERT_FILE
             HTTPS_PORT
             HTTPS_BIND_IP
             HTTPS_USE_IP_FROM_FORWARDED_HEADER
             HTTPS_ROTATION_PERIOD
             RECAPTCHA_ENABLED
             RECAPTCHA_PUB_KEY
             RECAPTCHA_SEC_KEY
             RECAPTCHA_REMOTEIP
             GIMP_CAPTCHA_ENABLED
             GIMP_CAPTCHA_DIR
             GIMP_CAPTCHA_HMAC_KEYFILE
             GIMP_CAPTCHA_RSA_KEYFILE
             SERVER_PUBLIC_FQDN
             CSP_ENABLED
             CSP_REPORT_ONLY
             CSP_INCLUDE_SELF
    :type distributor: :class:`bridgedb.https.distributor.HTTPSDistributor`
    :param distributor: A bridge distributor.
    :raises SystemExit: if the servers cannot be started.
    :rtype: :api:`twisted.web.server.Site`
    :returns: A webserver.
    """
    captcha = None
    fwdHeaders = config.HTTP_USE_IP_FROM_FORWARDED_HEADER
    numBridges = config.HTTPS_N_BRIDGES_PER_ANSWER
    fprInclude = config.HTTPS_INCLUDE_FINGERPRINTS

    logging.info("Starting web servers...")

    setFQDN(config.SERVER_PUBLIC_FQDN)

    index   = IndexResource()
    options = OptionsResource()
    howto   = HowtoResource()
    robots  = static.File(os.path.join(TEMPLATE_DIR, 'robots.txt'))
    assets  = static.File(os.path.join(TEMPLATE_DIR, 'assets/'))
    keys    = static.Data(bytes(strings.BRIDGEDB_OPENPGP_KEY), 'text/plain')
    csp     = CSPResource(enabled=config.CSP_ENABLED,
                          includeSelf=config.CSP_INCLUDE_SELF,
                          reportViolations=config.CSP_REPORT_ONLY,
                          useForwardedHeader=fwdHeaders)

    root = CustomErrorHandlingResource()
    root.putChild('', index)
    root.putChild('robots.txt', robots)
    root.putChild('keys', keys)
    root.putChild('assets', assets)
    root.putChild('options', options)
    root.putChild('howto', howto)
    root.putChild('maintenance', maintenance)
    root.putChild('error', resource500)
    root.putChild(CSPResource.reportURI, csp)

    if config.RECAPTCHA_ENABLED:
        publicKey = config.RECAPTCHA_PUB_KEY
        secretKey = config.RECAPTCHA_SEC_KEY
        captcha = partial(ReCaptchaProtectedResource,
                          remoteIP=config.RECAPTCHA_REMOTEIP)
    elif config.GIMP_CAPTCHA_ENABLED:
        # Get the master HMAC secret key for CAPTCHA challenges, and then
        # create a new HMAC key from it for use on the server.
        captchaKey = crypto.getKey(config.GIMP_CAPTCHA_HMAC_KEYFILE)
        hmacKey = crypto.getHMAC(captchaKey, "Captcha-Key")
        # Load or create our encryption keys:
        secretKey, publicKey = crypto.getRSAKey(config.GIMP_CAPTCHA_RSA_KEYFILE)
        captcha = partial(GimpCaptchaProtectedResource,
                          hmacKey=hmacKey,
                          captchaDir=config.GIMP_CAPTCHA_DIR)

    if config.HTTPS_ROTATION_PERIOD:
        count, period = config.HTTPS_ROTATION_PERIOD.split()
        sched = ScheduledInterval(count, period)
    else:
        sched = Unscheduled()

    bridges = BridgesResource(distributor, sched, numBridges, fwdHeaders,
                              includeFingerprints=fprInclude)
    if captcha:
        # Protect the 'bridges' page with a CAPTCHA, if configured to do so:
        protected = captcha(publicKey=publicKey,
                            secretKey=secretKey,
                            useForwardedHeader=fwdHeaders,
                            protectedResource=bridges)
        root.putChild('bridges', protected)
        logging.info("Protecting resources with %s." % captcha.func.__name__)
    else:
        root.putChild('bridges', bridges)

    site = Site(root)
    site.displayTracebacks = False

    if config.HTTP_UNENCRYPTED_PORT:  # pragma: no cover
        ip = config.HTTP_UNENCRYPTED_BIND_IP or ""
        port = config.HTTP_UNENCRYPTED_PORT or 80
        try:
            reactor.listenTCP(port, site, interface=ip)
        except CannotListenError as error:
            raise SystemExit(error)
        logging.info("Started HTTP server on %s:%d" % (str(ip), int(port)))

    if config.HTTPS_PORT:  # pragma: no cover
        ip = config.HTTPS_BIND_IP or ""
        port = config.HTTPS_PORT or 443
        try:
            from twisted.internet.ssl import DefaultOpenSSLContextFactory
            factory = DefaultOpenSSLContextFactory(config.HTTPS_KEY_FILE,
                                                   config.HTTPS_CERT_FILE)
            reactor.listenSSL(port, site, factory, interface=ip)
        except CannotListenError as error:
            raise SystemExit(error)
        logging.info("Started HTTPS server on %s:%d" % (str(ip), int(port)))

    return site



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_schedule -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Nick Mathewson
#           Isis Lovecruft <isis@torproject.org> 0xa3adb67a2cdb8b35
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2014-2015, Isis Lovecruft
# :license: see LICENSE for licensing information

"""This module implements functions for dividing time into chunks.

.. inheritance-diagram:: UnknownInterval Unscheduled ScheduledInterval
    :parts: 1
"""

import calendar

import math

from datetime import datetime

from zope import interface
from zope.interface import implementer
from zope.interface import Attribute


#: The known time intervals (or *periods*) for dividing time by.
KNOWN_INTERVALS = ["second", "minute", "hour", "day", "week", "month"]


class UnknownInterval(ValueError):
    """Raised if an interval isn't one of the :data:`KNOWN_INTERVALS`."""


def toUnixSeconds(timestruct):
    """Convert a datetime struct to a Unix timestamp in seconds.

    :type timestruct: :any:`datetime.datetime`
    :param timestruct: A ``datetime`` object to convert into a timestamp in
        Unix Era seconds.
    :rtype: int
    """
    return calendar.timegm(timestruct)

def fromUnixSeconds(timestamp):
    """Convert a Unix timestamp to a datetime struct.

    :param int timestamp: A timestamp in Unix Era seconds.
    :rtype: :any:`datetime.datetime`
    """
    return datetime.fromtimestamp(timestamp)


class ISchedule(interface.Interface):
    """An ``Interface`` specification for a Schedule."""

    intervalPeriod = Attribute(
        "The type of period which this Schedule's intervals will rotate by.")
    intervalCount = Attribute(
        "Number of **intervalPeriod**s before rotation to the next interval")

    def intervalStart(when=None):
        """Get the start time of the interval that contains **when**."""

    def getInterval(when=None):
        """Get the interval which includes an arbitrary **when**."""

    def nextIntervalStarts(when=None):
        """Get the start of the interval after the one containing **when**."""


@implementer(ISchedule)
class Unscheduled(object):
    """A base ``Schedule`` that has only one period that contains all time.

    >>> from bridgedb.schedule import fromUnixSeconds
    >>> from bridgedb.schedule import Unscheduled
    >>> timestamp = 1427769526
    >>> str(fromUnixSeconds(timestamp))
    '2015-03-31 02:38:46'
    >>> sched = Unscheduled()
    >>> start = sched.intervalStart(timestamp)
    >>> start
    -62135596800
    >>> str(fromUnixSeconds(start))
    '0001-01-01 00:00:00'
    >>> sched.getInterval(timestamp)
    '1970-01-01 00:00:00'
    >>> next = sched.nextIntervalStarts(timestamp)
    >>> next
    253402300799
    >>> str(fromUnixSeconds(next))
    '9999-12-31 23:59:59'

    """

    def __init__(self, count=None, period=None):
        """Create a schedule for dividing time into intervals.

        :param int count: The total number of **period** in one interval.
        :param str period: One of the periods in :data:`KNOWN_INTERVALS`.
        """
        self.intervalCount = count
        self.intervalPeriod = period

    def intervalStart(self, when=0):
        """Get the start time of the interval that contains **when**.

        :param int when: The time which we're trying to find the corresponding
            interval for.
        :rtype: int
        :returns: The Unix epoch timestamp for the start time of the interval
            that contains **when**.
        """
        return toUnixSeconds(datetime.min.timetuple())

    def getInterval(self, when=0):
        """Get the interval that contains the time **when**.

        .. note: We explicitly ignore the ``when`` parameter in this
            implementation because if something is Unscheduled then
            all timestamps should reside within the same period.

        :param int when: The time which we're trying to find the corresponding
            interval for.
        :rtype: str
        :returns: A timestamp in the form ``YEAR-MONTH[-DAY[-HOUR]]``. It's
            specificity depends on what type of interval we're using. For
            example, if using ``"month"``, the return value would be something
            like ``"2013-12"``.
        """
        return fromUnixSeconds(0).strftime('%04Y-%02m-%02d %02H:%02M:%02S')

    def nextIntervalStarts(self, when=0):
        """Return the start time of the interval starting _after_ when.

        :rtype: int
        :returns: Return the Y10K bug.
        """
        return toUnixSeconds(datetime.max.timetuple())


@implementer(ISchedule)
class ScheduledInterval(Unscheduled):
    """An class that splits time into periods, based on seconds, minutes,
    hours, days, weeks, or months.

    >>> from bridgedb.schedule import fromUnixSeconds
    >>> from bridgedb.schedule import ScheduledInterval
    >>> timestamp = 1427769526
    >>> str(fromUnixSeconds(timestamp))
    '2015-03-31 02:38:46'
    >>> sched = ScheduledInterval(5, 'minutes')
    >>> start = sched.intervalStart(timestamp)
    >>> start
    1427769300
    >>> current = sched.getInterval(timestamp)
    >>> current
    '2015-03-31 02:35:00'
    >>> current == str(fromUnixSeconds(start))
    True
    >>> next = sched.nextIntervalStarts(timestamp)
    >>> next
    1427769600
    >>> str(fromUnixSeconds(next))
    '2015-03-31 02:40:00'
    >>> later = 1427771057
    >>> str(fromUnixSeconds(later))
    '2015-03-31 03:04:17'
    >>> sched.getInterval(later)
    '2015-03-31 03:00:00'

    :ivar str intervalPeriod: One of the :data:`KNOWN_INTERVALS`.
    :ivar int intervalCount: The number of times :attr:`intervalPeriod` should
        be repeated within an interval.
    """

    def __init__(self, count=None, period=None):
        """Create a schedule for dividing time into intervals.

        :type count: :any:`int` or :any:`str`
        :param count: The total number of **period** in one interval.
        :param str period: One of the periods in :data:`KNOWN_INTERVALS`.
        """
        super(ScheduledInterval, self).__init__(count, period)
        self._setIntervalCount(count)
        self._setIntervalPeriod(period)

    def _setIntervalCount(self, count=None):
        """Set our :attr:`intervalCount`.

        .. attention:: This method should be called *before*
            :meth:`_setIntervalPeriod`, because the latter may change the
            count, if it decides to change the period (for example, to
            simplify things by changing weeks into days).

        :param int count: The number of times the :attr:`intervalPeriod`
            should be repeated during the interval. Defaults to ``1``.
        :raises UnknownInterval: if the specified **count** was invalid.
        """
        try:
            if not count > 0:
                count = 1
            count = int(count)
        except (TypeError, ValueError):
            raise UnknownInterval("%s.intervalCount: %r ist not an integer."
                                  % (self.__class__.__name__, count))
        self.intervalCount = count

    def _setIntervalPeriod(self, period=None):
        """Set our :attr:`intervalPeriod`.

        :param str period: One of the :data:`KNOWN_INTERVALS`, or its
            plural. Defaults to ``'hour'``.
        :raises UnknownInterval: if the specified **period** is unknown.
        """
        if not period:
            period = 'hour'
        try:
            period = period.lower()
            # Depluralise the period if necessary, i.e., "months" -> "month".
            if period.endswith('s'):
                period = period[:-1]

            if not period in KNOWN_INTERVALS:
                raise ValueError
        except (TypeError, AttributeError, ValueError):
            raise UnknownInterval("%s doesn't know about the %r interval type."
                                  % (self.__class__.__name__, period))
        self.intervalPeriod = period

        if period == 'week':
            self.intervalPeriod = 'day'
            self.intervalCount *= 7

    def intervalStart(self, when=0):
        """Get the start time of the interval that contains **when**.

        :param int when: The time which we're trying to determine the start of
            interval that contains it. This should be given in Unix seconds,
            for example, taken from :func:`calendar.timegm`.
        :rtype: int
        :returns: The Unix epoch timestamp for the start time of the interval
            that contains **when**.
        """
        # Convert `when`s which are floats, i.e. from time.time(), to ints:
        when = int(math.ceil(when))

        if self.intervalPeriod == 'month':
            # For months, we always start at the beginning of the month.
            date = fromUnixSeconds(when)
            months = (date.year * 12) + (date.month - 1)
            months -= (months % self.intervalCount)
            month = months % 12 + 1
            return toUnixSeconds((months // 12, month, 1, 0, 0, 0))
        elif self.intervalPeriod == 'day':
            # For days, we start at the beginning of a day.
            when -= when % (86400 * self.intervalCount)
            return when
        elif self.intervalPeriod == 'hour':
            # For hours, we start at the beginning of an hour.
            when -= when % (3600 * self.intervalCount)
            return when
        elif self.intervalPeriod == 'minute':
            when -= when % (60 * self.intervalCount)
            return when
        elif self.intervalPeriod == 'second':
            when -= when % self.intervalCount
            return when

    def getInterval(self, when=0):
        """Get the interval that contains the time **when**.

        >>> import calendar
        >>> from bridgedb.schedule import ScheduledInterval
        >>> sched = ScheduledInterval(1, 'month')
        >>> when = calendar.timegm((2007, 12, 12, 0, 0, 0))
        >>> sched.getInterval(when)
        '2007-12'
        >>> then = calendar.timegm((2014, 05, 13, 20, 25, 13))
        >>> sched.getInterval(then)
        '2014-05'

        :param int when: The time which we're trying to find the corresponding
            interval for. Given in Unix seconds, for example, taken from
            :func:`calendar.timegm`.
        :rtype: str
        :returns: A timestamp in the form ``YEAR-MONTH[-DAY[-HOUR]]``. It's
            specificity depends on what type of interval we're using. For
            example, if using ``"month"``, the return value would be something
            like ``"2013-12"``.
        """
        date = fromUnixSeconds(self.intervalStart(when))

        fstr = "%04Y-%02m"
        if self.intervalPeriod != 'month':
            fstr += "-%02d"
            if self.intervalPeriod != 'day':
                fstr += " %02H"
                if self.intervalPeriod != 'hour':
                    fstr += ":%02M"
                    if self.intervalPeriod == 'minute':
                        fstr += ":%02S"

        return date.strftime(fstr)

    def nextIntervalStarts(self, when=0):
        """Return the start time of the interval starting _after_ when.

        :returns: The Unix epoch timestamp for the start time of the interval
            that contains **when**.
        """
        seconds = self.intervalStart(when)

        if self.intervalPeriod == 'month':
            date = fromUnixSeconds(seconds)
            year = date.year
            months = date.month + self.intervalCount
            if months > 12:
                year = date.year + 1
                months = months - 12
            return toUnixSeconds((year, months, 1, 0, 0, 0))
        elif self.intervalPeriod == 'day':
            return seconds + (86400 * self.intervalCount)
        elif self.intervalPeriod == 'hour':
            return seconds + (3600 * self.intervalCount)
        elif self.intervalPeriod == 'minute':
            return seconds + (60 * self.intervalCount)
        elif self.intervalPeriod == 'second':
            return seconds + self.intervalCount



# -*- coding: utf-8; test-case-name: bridgedb.test.test_email_request; -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Nick Mathewson <nickm@torproject.org>
#           Isis Lovecruft <isis@torproject.org> 0xA3ADB67A2CDB8B35
#           Matthew Finkel <sysrqb@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2013-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

"""
.. py:module:: bridgedb.email.request
    :synopsis: Classes for parsing and storing information about requests for
               bridges which are sent to the email distributor.

bridgedb.email.request
======================

Classes for parsing and storing information about requests for bridges
which are sent to the email distributor.

.. inheritance-diagram:: EmailBridgeRequest
    :parts: 1

::

  bridgedb.email.request
   | |_ determineBridgeRequestOptions - Figure out which filters to apply, or
   |                                    offer help.
   |_ EmailBridgeRequest - A request for bridges which was received through
                           the email distributor.

..
"""

from __future__ import print_function
from __future__ import unicode_literals

import logging
import re

from bridgedb import bridgerequest
from bridgedb.email.distributor import EmailRequestedHelp
from bridgedb.email.distributor import EmailRequestedKey


#: A regular expression for matching the Pluggable Transport method TYPE in
#: emailed requests for Pluggable Transports.
TRANSPORT_REGEXP = ".*transport ([a-z][_a-z0-9]*)"
TRANSPORT_PATTERN = re.compile(TRANSPORT_REGEXP)

#: A regular expression that matches country codes in requests for unblocked
#: bridges.
UNBLOCKED_REGEXP = ".*unblocked ([a-z]{2,4})"
UNBLOCKED_PATTERN = re.compile(UNBLOCKED_REGEXP)


def determineBridgeRequestOptions(lines):
    """Figure out which :mod:`~bridgedb.filters` to apply, or offer help.

    .. note:: If any ``'transport TYPE'`` was requested, or bridges not
        blocked in a specific CC (``'unblocked CC'``), then the ``TYPE``
        and/or ``CC`` will *always* be stored as a *lowercase* string.

    :param list lines: A list of lines from an email, including the headers.
    :raises EmailRequestedHelp: if the client requested help.
    :raises EmailRequestedKey: if the client requested our GnuPG key.
    :rtype: :class:`EmailBridgeRequest`
    :returns: A :class:`~bridgerequest.BridgeRequest` with all of the requested
        parameters set. The returned ``BridgeRequest`` will have already had
        its filters generated via :meth:`~EmailBridgeRequest.generateFilters`.
    """
    request = EmailBridgeRequest()
    skippedHeaders = False

    for line in lines:
        line = line.strip().lower()
        # Ignore all lines before the first empty line:
        if not line: skippedHeaders = True
        if not skippedHeaders: continue

        if ("help" in line) or ("halp" in line):
            raise EmailRequestedHelp("Client requested help.")

        if "get" in line:
            request.isValid(True)
            logging.debug("Email request was valid.")
        if "key" in line:
            request.wantsKey(True)
            raise EmailRequestedKey("Email requested a copy of our GnuPG key.")
        if "ipv6" in line:
            request.withIPv6()
        if "transport" in line:
            request.withPluggableTransportType(line)
        if "unblocked" in line:
            request.withoutBlockInCountry(line)

    logging.debug("Generating hashring filters for request.")
    request.generateFilters()
    return request


class EmailBridgeRequest(bridgerequest.BridgeRequestBase):
    """We received a request for bridges through the email distributor."""

    def __init__(self):
        """Process a new bridge request received through the
        :class:`~bridgedb.email.distributor.EmailDistributor`.
        """
        super(EmailBridgeRequest, self).__init__()
        self._wantsKey = False

    def wantsKey(self, wantsKey=None):
        """Get or set whether this bridge request wanted our GnuPG key.

        If called without parameters, this method will return the current
        state, otherwise (if called with the **wantsKey** parameter set), it
        will set the current state for whether or not this request wanted our
        key.

        :param bool wantsKey: If given, set the validity state of this
            request. Otherwise, get the current state.
        """
        if wantsKey is not None:
            self._wantsKey = bool(wantsKey)
        return self._wantsKey

    def withoutBlockInCountry(self, line):
        """This request was for bridges not blocked in **country**.

        Add any country code found in the **line** to the list of
        ``notBlockedIn``. Currently, a request for a transport is recognized
        if the email line contains the ``'unblocked'`` command.

        :param str country: The line from the email wherein the client
            requested some type of Pluggable Transport.
        """
        unblocked = None

        logging.debug("Parsing 'unblocked' line: %r" % line)
        try:
            unblocked = UNBLOCKED_PATTERN.match(line).group(1)
        except (TypeError, AttributeError):
            pass

        if unblocked:
            self.notBlockedIn.append(unblocked)
            logging.info("Email requested bridges not blocked in: %r"
                         % unblocked)

    def withPluggableTransportType(self, line):
        """This request included a specific Pluggable Transport identifier.

        Add any Pluggable Transport method TYPE found in the **line** to the
        list of ``transports``. Currently, a request for a transport is
        recognized if the email line contains the ``'transport'`` command.

        :param str line: The line from the email wherein the client
            requested some type of Pluggable Transport.
        """
        transport = None
        logging.debug("Parsing 'transport' line: %r" % line)

        try:
            transport = TRANSPORT_PATTERN.match(line).group(1)
        except (TypeError, AttributeError):
            pass

        if transport:
            self.transports.append(transport)
            logging.info("Email requested transport type: %r" % transport)



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_qrcodes ; -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2014-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

"""Utilities for working with QRCodes."""


import cStringIO
import logging

try:
    import qrcode
except ImportError:  # pragma: no cover
    qrcode = False
    logging.warn("Could not import Python qrcode module.")
    logging.debug(("You'll need the qrcode Python module for this to "
                   "work. On Debian-based systems, this should be in the "
                   "python-qrcode package."))


def generateQR(bridgelines, imageFormat=u'JPEG', bridgeSchema=False):
    """Generate a QRCode for the client's bridge lines.

    :param str bridgelines: The Bridge Lines which we are distributing to the
        client.
    :param bool bridgeSchema: If ``True``, prepend ``'bridge://'`` to the
        beginning of each bridge line before QR encoding.
    :rtype: str or ``None``
    :returns: The generated QRCode, as a string.
    """
    logging.debug("Attempting to encode bridge lines into a QRCode...")

    if not bridgelines:
        return

    if not qrcode:
        logging.info("Not creating QRCode for bridgelines; no qrcode module.")
        return

    try:
        if bridgeSchema:
            # See https://bugs.torproject.org/12639 for why bridge:// is used.
            # (Hopefully, Orbot will pick up the ACTION_VIEW intent.)
            schema = 'bridge://'
            prefixed = []
            for line in bridgelines.strip().split('\n'):
                prefixed.append(schema + line)
            bridgelines = '\n'.join(prefixed)

        logging.debug("QR encoding bridge lines: %s" % bridgelines)

        qr = qrcode.QRCode()
        qr.add_data(bridgelines)

        buf = cStringIO.StringIO()
        img = qr.make_image().resize([350, 350])
        img.save(buf, imageFormat)
        buf.seek(0)

        imgstr = buf.read()
        return imgstr

    except KeyError as error:
        logging.error(str(error))
        logging.debug(("It seems python-imaging doesn't understand how to "
                       "save in the %s format.") % imageFormat)
    except Exception as error:  # pragma: no cover
        logging.error(("There was an error while attempting to generate the "
                       "QRCode: %s") % str(error))



# -*- coding: utf-8 -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2014-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________


from twisted.trial import unittest

from bridgedb import bridgerequest
from bridgedb.bridgerequest import IRequestBridges
from bridgedb.bridgerequest import BridgeRequestBase


class BridgeRequestBaseTests(unittest.TestCase):
    """Unittests for :class:`bridgedb.bridgerequest.BridgeRequestBase`."""

    def setUp(self):
        """Setup test run."""
        self.request = BridgeRequestBase()

    def test_BridgeRequestBase_implements_IRequestBridges(self):
        """BridgeRequestBase should implement IRequestBridges interface."""
        self.assertTrue(IRequestBridges.implementedBy(BridgeRequestBase))

    def test_BridgeRequestBase_withoutBlockInCountry(self):
        """BridgeRequestBase.withoutBlockInCountry() should add the country CC
        to the ``notBlockedIn`` attribute.
        """
        self.request.withoutBlockInCountry('US')
        self.assertIn('us', self.request.notBlockedIn)

    def test_BridgeRequestBase_withPluggableTransportType(self):
        """BridgeRequestBase.withPluggableTransportType() should add the
        pluggable transport type to the ``transport`` attribute.
        """
        self.request.withPluggableTransportType('huggable_transport')
        self.assertIn('huggable_transport', self.request.transports)

    def test_BridgeRequestBase_getHashringPlacement_without_client(self):
        """BridgeRequestBase.getHashringPlacement() without a client parameter
        should use the default client identifier string.
        """
        self.assertEqual(self.request.getHashringPlacement('AAAA'),
                         3486762050L)

    def test_BridgeRequestBase_getHashringPlacement_with_client(self):
        """BridgeRequestBase.getHashringPlacement() with a client parameter
        should use the client identifier string.
        """
        self.assertEqual(self.request.getHashringPlacement('AAAA', client='you'),
                         2870307088L)



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
# :copyright: (c) 2014-2015, Isis Lovecruft
#             (c) 2014-2015, The Tor Project, Inc.
# :license: see LICENSE for licensing information

"""Unittests for the :mod:`bridgedb.schedule` module."""

from __future__ import print_function

from twisted.trial import unittest

from bridgedb import schedule


class UnscheduledTests(unittest.TestCase):
    """Tests for :class:`bridgedb.scheduled.Unscheduled`."""

    def setUp(self):
        self.sched = schedule.Unscheduled()

    def test_Unscheduled_init(self):
        """The instance should be an instance of its class."""
        self.assertIsInstance(self.sched, schedule.Unscheduled)

    def test_Unscheduled_providesISchedule(self):
        """Unscheduled should implement the ISchedule interface."""
        schedule.ISchedule.namesAndDescriptions()
        self.assertTrue(schedule.ISchedule.providedBy(self.sched))

    def test_Unscheduled_intervalStart_noargs(self):
        time = self.sched.intervalStart()
        self.assertIsInstance(time, int)
        self.assertEquals(time, -62135596800)

    def test_Unscheduled_getInterval_is_constant(self):
        import time
        now = time.time()

        interval_default = self.sched.getInterval()
        self.assertIsInstance(interval_default, str)

        interval_zero = self.sched.getInterval(0)
        self.assertIsInstance(interval_zero, str)

        interval_now = self.sched.getInterval(now)
        self.assertIsInstance(interval_now, str)

        self.assertEquals(interval_default, interval_zero)
        self.assertEquals(interval_default, interval_now)

    def test_Unscheduled_nextIntervalStarts_noargs(self):
        time = self.sched.nextIntervalStarts()
        self.assertIsInstance(time, int)
        self.assertEquals(time, 253402300799)


class ScheduledIntervalTests(unittest.TestCase):
    """Tests for :class:`bridgedb.scheduled.ScheduledInterval`."""

    def setUp(self):
        import time
        self.now = time.time
        self.sched = schedule.ScheduledInterval

    def test_ScheduledInterval_providesISchedule(self):
        """ScheduledInterval should implement the ISchedule interface."""
        self.assertTrue(schedule.ISchedule.providedBy(self.sched(1, 'month')))

    def _check_init(self, sched):
        """The instance should be an instance of its class."""
        self.assertIsInstance(sched, schedule.ScheduledInterval)

    def test_ScheduledInterval_init_month(self):
        self._check_init(self.sched(1, 'month'))

    def test_ScheduledInterval_init_week(self):
        self._check_init(self.sched(2, 'week'))

    def test_ScheduledInterval_init_day(self):
        self._check_init(self.sched(5, 'days'))

    def test_ScheduledInterval_init_hour(self):
        self._check_init(self.sched(12, 'hours'))

    def test_ScheduledInterval_init_minute(self):
        self._check_init(self.sched(10, 'minute'))

    def test_ScheduledInterval_init_seconds(self):
        self._check_init(self.sched(30, 'seconds'))

    def test_ScheduledInterval_init_badIntervalPeriod(self):
        self.assertRaises(schedule.UnknownInterval,
                          self.sched, 2, 'decades')

    def test_ScheduledInterval_init_badIntervalCount(self):
        self.assertRaises(schedule.UnknownInterval,
                          self.sched, 'd20', 'minutes')

    def test_ScheduledInterval_init_negativeIntervalCount(self):
        sched = self.sched(-100000, 'days')
        self.assertEquals(sched.intervalCount, 1)
        self.assertEquals(sched.intervalPeriod, 'day')

    def test_ScheduledInterval_init_noargs(self):
        """Check that the defaults parameters function as expected."""
        sched = self.sched()
        self.assertEquals(sched.intervalCount, 1)
        self.assertEquals(sched.intervalPeriod, 'hour')

    def _check_intervalStart(self, count=30, period='second', variance=30):
        """Test the ScheduledInterval.intervalStart() method.

        :param int count: The number of **period**s within an interval.
        :param str period: The interval type for the period.
        :param int variance: The amount of variance (in seconds) to tolerate
            between the start of the interval containing now, and now.
        """
        now = int(self.now())
        sched = self.sched(count, period)
        time = sched.intervalStart(now)
        self.assertIsInstance(time, int)
        self.assertApproximates(now, time, variance)

    def test_ScheduledInterval_intervalStart_month(self):
        self._check_intervalStart(1, 'month', 31*24*60*60)

    def test_ScheduledInterval_intervalStart_week(self):
        self._check_intervalStart(2, 'week', 14*24*60*60)

    def test_ScheduledInterval_intervalStart_day(self):
        self._check_intervalStart(5, 'days', 5*24*60*60)

    def test_ScheduledInterval_intervalStart_hour(self):
        self._check_intervalStart(12, 'hours', 12*60*60)

    def test_ScheduledInterval_intervalStart_minute(self):
        self._check_intervalStart(10, 'minute', 10*60)

    def test_ScheduledInterval_intervalStart_seconds(self):
        self._check_intervalStart(30, 'seconds', 30)

    def test_ScheduledInterval_intervalStart_time_time(self):
        """Calling ScheduledInterval.intervalStart(time.time()) should only
        return ints, not floats.
        """
        import time

        timestamp = time.time()
        sched = self.sched(5, 'minutes')

        self.assertIsInstance(timestamp, float)
        self.assertIsInstance(sched.intervalStart(timestamp), int)

    def _check_getInterval(self, count=30, period='second', variance=30):
        """Test the ScheduledInterval.getInterval() method.

        :param int count: The number of **period**s within an interval.
        :param str period: The interval type for the period.
        :param int variance: The amount of variance (in seconds) to tolerate
            between the start of the interval containing now, and now.
        """
        now = int(self.now())
        sched = self.sched(count, period)
        ts = sched.getInterval(now)
        self.assertIsInstance(ts, str)
        secs = [int(x) for x in ts.replace('-', ' ').replace(':', ' ').split()]
        [secs.append(0) for _ in xrange(6-len(secs))]
        secs = schedule.calendar.timegm(secs)
        self.assertApproximates(now, secs, variance)

    def test_ScheduledInterval_getInterval_month(self):
        self._check_getInterval(2, 'month', 2*31*24*60*60)

    def test_ScheduledInterval_getInterval_week(self):
        self._check_getInterval(1, 'week', 7*24*60*60)

    def test_ScheduledInterval_getInterval_day(self):
        self._check_getInterval(4, 'days', 4*24*60*60)

    def test_ScheduledInterval_getInterval_hour(self):
        self._check_getInterval(23, 'hours', 23*60*60)

    def test_ScheduledInterval_getInterval_minute(self):
        self._check_getInterval(15, 'minutes', 15*60)

    def test_ScheduledInterval_getInterval_seconds(self):
        self._check_getInterval(10, 'seconds', 60)

    def _check_nextIntervalStarts(self, count=30, period='second', variance=30):
        """Test the ScheduledInterval.nextIntervalStarts() method.

        :param int count: The number of **period**s within an interval.
        :param str period: The interval type for the period.
        :param int variance: The amount of variance (in seconds) to tolerate
            between the start of the interval containing now, and now.
        """
        now = int(self.now())
        sched = self.sched(count, period)
        time = sched.nextIntervalStarts(now)
        self.assertIsInstance(time, int)
        # (now + variance - time) should be > variance
        self.assertApproximates(now + variance, time, variance)

    def test_ScheduledInterval_nextIntervalStarts_month(self):
        self._check_nextIntervalStarts(2, 'month', 2*31*24*60*60)

    def test_ScheduledInterval_nextIntervalStarts_week(self):
        self._check_nextIntervalStarts(1, 'week', 7*24*60*60)

    def test_ScheduledInterval_nextIntervalStarts_day(self):
        self._check_nextIntervalStarts(4, 'days', 4*24*60*60)

    def test_ScheduledInterval_nextIntervalStarts_hour(self):
        self._check_nextIntervalStarts(23, 'hours', 23*60*60)

    def test_ScheduledInterval_nextIntervalStarts_minute(self):
        self._check_nextIntervalStarts(15, 'minutes', 15*60)

    def test_ScheduledInterval_nextIntervalStarts_seconds(self):
        self._check_nextIntervalStarts(10, 'seconds', 10)



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_email_distributor -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Nick Mathewson
#           Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           Matthew Finkel 0x017DD169EA793BE2 <sysrqb@torproject.org>
# :copyright: (c) 2013-2015, Isis Lovecruft
#             (c) 2013-2015, Matthew Finkel
#             (c) 2007-2015, The Tor Project, Inc.
# :license: see LICENSE for licensing information

"""
.. py:module:: bridgedb.email.distributor
    :synopsis: A Distributor which hands out Bridges via an email interface.

bridgedb.email.autoresponder
============================

A :class:`~bridgedb.distribute.Distributor` which hands out :class:`bridges
<bridgedb.bridges.Bridge>` to clients via an email interface.

.. inheritance-diagram:: IgnoreEmail TooSoonEmail EmailRequestedHelp EmailRequestedKey EmailDistributor
    :parts: 1
"""

import logging
import time

import bridgedb.Storage

from bridgedb.Bridges import BridgeRing
from bridgedb.Bridges import FilteredBridgeSplitter
from bridgedb.crypto import getHMAC
from bridgedb.crypto import getHMACFunc
from bridgedb.distribute import Distributor
from bridgedb.filters import byFilters
from bridgedb.filters import byIPv4
from bridgedb.filters import byIPv6
from bridgedb.filters import bySubring
from bridgedb.parse import addr


#: The minimum amount of time (in seconds) which must pass before a client who
#: has previously been given an email response must wait before being eligible
#: to receive another response.
MAX_EMAIL_RATE = 3 * 3600


class IgnoreEmail(addr.BadEmail):
    """Raised when we get requests from this address after rate warning."""


class TooSoonEmail(addr.BadEmail):
    """Raised when we got a request from this address too recently."""


class EmailRequestedHelp(Exception):
    """Raised when a client has emailed requesting help."""


class EmailRequestedKey(Exception):
    """Raised when an incoming email requested a copy of our GnuPG keys."""


class EmailDistributor(Distributor):
    """Object that hands out bridges based on the email address of an incoming
    request and the current time period.

    :type hashring: :class:`~bridgedb.Bridges.BridgeRing`
    :ivar hashring: A hashring to hold all the bridges we hand out.
    """

    #: The minimum amount of time (in seconds) which must pass before a client
    #: who has previously been given an email response must wait before being
    #: eligible to receive another response.
    emailRateMax = MAX_EMAIL_RATE

    def __init__(self, key, domainmap, domainrules,
                 answerParameters=None, whitelist=None):
        """Create a bridge distributor which uses email.

        :type emailHmac: callable
        :param emailHmac: An hmac function used to order email addresses
            within a ring. See :func:`~bridgedb.crypto.getHMACFunc`.
        :param dict domainmap: A map from lowercase domains that we support
            mail from to their canonical forms. See `EMAIL_DOMAIN_MAP` option
            in `bridgedb.conf`.
        :param domainrules: DOCDOC
        :param answerParameters: DOCDOC
        :type whitelist: dict or ``None``
        :param whitelist: A dictionary that maps whitelisted email addresses
            to GnuPG fingerprints.
        """
        super(EmailDistributor, self).__init__(key)

        self.domainmap = domainmap
        self.domainrules = domainrules
        self.whitelist = whitelist or dict()
        self.answerParameters = answerParameters

        key1 = getHMAC(key, "Map-Addresses-To-Ring")
        key2 = getHMAC(key, "Order-Bridges-In-Ring")

        self.emailHmac = getHMACFunc(key1, hex=False)
        #XXX cache options not implemented
        self.hashring = FilteredBridgeSplitter(key2, max_cached_rings=5)

        self.name = "Email"

    def bridgesPerResponse(self, hashring=None):
        return super(EmailDistributor, self).bridgesPerResponse(hashring)

    def getBridges(self, bridgeRequest, interval, clock=None):
        """Return a list of bridges to give to a user.

        .. hint:: All checks on the email address (which should be stored in
            the ``bridgeRequest.client`` attribute), such as checks for
            whitelisting and canonicalization of domain name, are done in
            :meth:`bridgedb.email.autoresponder.getMailTo` and
            :meth:`bridgedb.email.autoresponder.SMTPAutoresponder.runChecks`.

        :type bridgeRequest:
            :class:`~bridgedb.email.request.EmailBridgeRequest`
        :param bridgeRequest: A
            :class:`~bridgedb.bridgerequest.BridgeRequestBase` with the
            :data:`~bridgedb.bridgerequest.BridgeRequestBase.client` attribute
            set to a string containing the client's full, canonicalized email
            address.
        :type interval: str
        :param interval: The time period when we got this request. This can be
            any string, so long as it changes with every period.
        :type clock: :api:`twisted.internet.task.Clock`
        :param clock: If given, use the clock to ask what time it is, rather
            than :api:`time.time`. This should likely only be used for
            testing.
        :rtype: :any:`list` or ``None``
        :returns: A list of :class:`~bridgedb.bridges.Bridges` for the
            ``bridgeRequest.client``, if allowed.  Otherwise, returns ``None``.
        """
        if (not bridgeRequest.client) or (bridgeRequest.client == 'default'):
            raise addr.BadEmail(
                ("%s distributor can't get bridges for invalid email address: "
                 "%s") % (self.name, bridgeRequest.client), bridgeRequest.client)

        logging.info("Attempting to get bridges for %s..." % bridgeRequest.client)

        now = time.time()

        if clock:
            now = clock.seconds()

        with bridgedb.Storage.getDB() as db:
            wasWarned = db.getWarnedEmail(bridgeRequest.client)
            lastSaw = db.getEmailTime(bridgeRequest.client)
            if lastSaw is not None:
                if bridgeRequest.client in self.whitelist:
                    logging.info(
                        "Whitelisted address %s was last seen %d seconds ago."
                        % (bridgeRequest.client, now - lastSaw))
                elif (lastSaw + self.emailRateMax) >= now:
                    wait = (lastSaw + self.emailRateMax) - now
                    logging.info("Client %s must wait another %d seconds."
                                 % (bridgeRequest.client, wait))
                    if wasWarned:
                        raise IgnoreEmail(
                            "Client %s was warned." % bridgeRequest.client,
                            bridgeRequest.client)
                    else:
                        logging.info("Sending duplicate request warning.")
                        db.setWarnedEmail(bridgeRequest.client, True, now)
                        db.commit()
                        raise TooSoonEmail("Must wait %d seconds" % wait,
                                           bridgeRequest.client)
            # warning period is over
            elif wasWarned:
                db.setWarnedEmail(bridgeRequest.client, False)

            pos = self.emailHmac("<%s>%s" % (interval, bridgeRequest.client))

            ring = None
            filtres = frozenset(bridgeRequest.filters)
            if filtres in self.hashring.filterRings:
                logging.debug("Cache hit %s" % filtres)
                _, ring = self.hashring.filterRings[filtres]
            else:
                logging.debug("Cache miss %s" % filtres)
                key = getHMAC(self.key, "Order-Bridges-In-Ring")
                ring = BridgeRing(key, self.answerParameters)
                self.hashring.addRing(ring, filtres, byFilters(filtres),
                                      populate_from=self.hashring.bridges)

            returnNum = self.bridgesPerResponse(ring)
            result = ring.getBridges(pos, returnNum)

            db.setEmailTime(bridgeRequest.client, now)
            db.commit()

        return result

    def cleanDatabase(self):
        """Clear all emailed response and warning times from the database."""
        logging.info(("Cleaning all response and warning times for the %s "
                      "distributor from the database...") % self.name)
        with bridgedb.Storage.getDB() as db:
            try:
                db.cleanEmailedBridges(time.time() - self.emailRateMax)
                db.cleanWarnedEmails(time.time() - self.emailRateMax)
            except:
                db.rollback()
                raise
            else:
                db.commit()

    def prepopulateRings(self):
        """Prepopulate this distributor's hashrings and subhashrings with
        bridges.
        """
        logging.info("Prepopulating %s distributor hashrings..." % self.name)

        for filterFn in [byIPv4, byIPv6]:
            ruleset = frozenset([filterFn])
            key = getHMAC(self.key, "Order-Bridges-In-Ring")
            ring = BridgeRing(key, self.answerParameters)
            self.hashring.addRing(ring, ruleset, byFilters([filterFn]),
                                  populate_from=self.hashring.bridges)

        # Since prepopulateRings is called every half hour when the bridge
        # descriptors are re-parsed, we should clean the database then.
        self.cleanDatabase()



# -*- coding: utf-8 -*-

"""
This module is responsible for everything concerning file bucket bridge 
distribution. File bucket bridge distribution means that unallocated bridges 
are allocated to a certain pseudo-distributor and later written to a file.

For example, the following is a dict of pseudo-distributors (also called 
'bucket identifiers') with numbers of bridges assigned to them:

        FILE_BUCKETS = { "name1": 10, "name2": 15, "foobar": 3 }

This configuration for buckets would result in 3 files being created for bridge
distribution: name1-2010-07-17.brdgs, name2-2010-07-17.brdgs and 
foobar-2010-07-17.brdgs. The first file would contain 10 bridges from BridgeDB's
'unallocated' pool. The second file would contain 15 bridges from the same pool
and the third one similarly 3 bridges. These files can then be handed out to 
trusted parties via mail or fed to other distribution mechanisms such as 
twitter.

Note that in BridgeDB slang, the _distributor_ would still be 'unallocated',
even though in the database, there would now by 'name1', 'name2' or 'foobar'
instead of 'unallocated'. This is why they are called pseudo-distributors.
"""

import logging
import time
import bridgedb.Storage
import bridgedb.Bridges 
import binascii
import sqlite3
from gettext import gettext as _
toHex = binascii.b2a_hex


# What should pseudo distributors be prefixed with in the database so we can
# distinguish them from real distributors?
PSEUDO_DISTRI_PREFIX = "pseudo_"

# Set to rediculously high number
BUCKET_MAX_BRIDGES = 1000000


class BucketData(object):
    """Configures a bridge bucket with the number of bridges which should be
    allocated, the name of the bucket, and other similar data.

    :param str name: The name of this bucket (from the config file). This will
        be prefixed by the :data:`PSEUDO_DISTRIBUTOR_PREFIX`.
    :type needed: str or int
    :param needed: The number of bridges needed for this bucket (also from the
        config file).
    :param int allocated: Number of bridges already allocated for this bucket.
    """
    def __init__(self, name, needed):
        self.name = name
        if needed == "*":
            needed = BUCKET_MAX_BRIDGES
        self.needed = int(needed)
        self.allocated = 0

class BucketManager(object):
    """BucketManager reads a number of file bucket identifiers from the config.

    They're expected to be in the following format::

        FILE_BUCKETS = { "name1": 10, "name2": 15, "foobar": 3 }

    This syntax means that certain buckets ("name1", "name2" and so on) are
    given a number of bridges (10, 15 and so on). Names can be anything.  The
    name will later be the prefix of the file that is written with the
    assigned number of bridges in it. Instead of a number, a wildcard item
    ("*") is allowed, too. This means that the corresponsing bucket file will
    get the maximum number of possible bridges (as many as are left in the
    unallocated bucket).

    The files will be written in ip:port format, one bridge per line.

    The way this works internally is as follows:

    First of all, the assignBridgesToBuckets() routine runs through the
    database of bridges and looks up the 'distributor' field of each
    bridge. Unallocated bridges are sent to a pool for later assignement.
    Already allocated bridges for file bucket distribution are sorted and
    checked.  They're checked for whether their bucket identifier still exists
    in the current config and also whether the number of assigned bridges is
    still valid. If either the bucket identifier is not existing anymore or
    too many bridges are currently assigned to it, bridges will go to the
    unassigned pool.

    In the second step, after bridges are sorted and the unassigned pool is
    ready, the assignBridgesToBuckets() routine assigns one bridge from the
    unassigned pool to a known bucket identifier at a time until it either
    runs out of bridges in the unallocated pool or the number of needed
    bridges for that bucket is reached.

    When all bridges are assigned in this way, they can then be dumped into
    files by calling the dumpBridges() routine.

    :type cfg: :class:`bridgedb.persistent.Conf`
    :ivar cfg: The central configuration instance.
    :ivar list bucketList: A list of BucketData instances, holding all
        configured (and thus requested) buckets with their respective numbers.
    :ivar list unallocatedList: Holds all bridges from the 'unallocated' pool.
    :ivar bool unallocated_available: Is at least one unallocated bridge
        available?
    :ivar str distributor_prefix: The 'distributor' field in the database will
        hold the name of our pseudo-distributor, prefixed by this string. By
        default, this uses :data:`PSEUDO_DISTRIBUTOR_PREFIX`.
    :ivar db: The bridge database instance.
    """

    def __init__(self, cfg):
        """Create a ``BucketManager``.

        :type cfg: :class:`bridgedb.persistent.Conf`
        :param cfg: The central configuration instance.
        """
        self.cfg = cfg
        self.bucketList = []
        self.unallocatedList = []
        self.unallocated_available = False
        self.distributor_prefix = PSEUDO_DISTRI_PREFIX

    def addToUnallocatedList(self, hex_key):
        """Add a bridge by **hex_key** into the unallocated pool."""
        with bridgedb.Storage.getDB() as db:
            try:
                db.updateDistributorForHexKey("unallocated", hex_key)
            except:
                db.rollback()
                raise
            else:
                db.commit()
        self.unallocatedList.append(hex_key)
        self.unallocated_available = True

    def getBucketByIdent(self, bucketIdent):
        """If we know this bucket identifier, then return the corresponding
        :class:`BucketData` object.
        """
        for d in self.bucketList:
            if d.name == bucketIdent:
                return d
        return None

    def assignUnallocatedBridge(self, bucket):
        """Assign an unallocated bridge to a certain **bucket**."""
        hex_key = self.unallocatedList.pop()
        # Mark pseudo-allocators in the database as such
        allocator_name = bucket.name
        #print "KEY: %d NAME: %s" % (hex_key, allocator_name)
        logging.debug("Moving %s to %s" % (hex_key, allocator_name))
        with bridgedb.Storage.getDB() as db:
            try:
                db.updateDistributorForHexKey(allocator_name, hex_key)
            except:
                db.rollback()
                logging.warn("Failed to move %s to new distributor (%s)"
                             % (hex_key, allocator_name))

                # Ok, this seems useless, but for consistancy's sake, we'll
                # re-assign the bridge from this missed db update attempt to the
                # unallocated list. Remember? We pop()'d it before.
                self.addToUnallocatedList(hex_key)
                raise
            else:
                db.commit()
        bucket.allocated += 1
        if len(self.unallocatedList) < 1:
            self.unallocated_available = False
        return True

    def assignBridgesToBuckets(self):
        """Read file bucket identifiers from the configuration, sort them, and
        write necessary changes to the database.
        """
        logging.debug("Assigning bridges to buckets for pseudo-distributors")
        # Build distributor list
        for k, v in self.cfg.FILE_BUCKETS.items():
            prefixed_key = self.distributor_prefix + k
            d = BucketData(prefixed_key, v)
            self.bucketList.append(d)

        # Loop through all bridges and sort out distributors
        with bridgedb.Storage.getDB() as db:
            allBridges = db.getAllBridges()
        for bridge in allBridges:
            if bridge.distributor == "unallocated":
                self.addToUnallocatedList(bridge.hex_key)
                continue

            # Filter non-pseudo distributors (like 'https' and 'email') early,
            # too
            if not bridge.distributor.startswith(self.distributor_prefix):
                continue

            # Return the bucket in case we know it already
            d = self.getBucketByIdent(bridge.distributor)
            if d is not None:
                # Does this distributor need another bridge? If not, re-inject
                # it into the 'unallocated' pool for for later assignment
                if d.allocated < d.needed:
                    d.allocated += 1
                else:
                    # Bucket has enough members already, free this one
                    self.addToUnallocatedList(bridge.hex_key)
            # We don't know it. Maybe an old entry. Free it.
            else:
                self.addToUnallocatedList(bridge.hex_key)

        # Loop through bucketList while we have and need unallocated
        # bridges, assign one bridge at a time
        while self.unallocated_available and len(self.bucketList) > 0:
            logging.debug("We have %d unallocated bridges and %d buckets to " \
                          "fill. Let's do it."
                          % (len(self.unallocatedList), len(self.bucketList)))
            for d in self.bucketList:
                if d.allocated < d.needed:
                    try:
                        if not self.assignUnallocatedBridge(d):
                            break
                    except sqlite3.DatabaseError as e:
                        dist = d.name.replace(self.distributor_prefix, "")
                        logging.warn("Couldn't assign unallocated bridge to " \
                                     "%s: %s" % (dist, e))
                else:
                    # When we have enough bridges, remove bucket identifier 
                    # from list
                    self.bucketList.remove(d)

    def dumpBridgesToFile(self, filename, bridges):
        """Dump a list of given **bridges** into **filename**."""
        logging.debug("Dumping bridge assignments to file: %r" % filename)
        # get the bridge histories and sort by Time On Same Address
        bridgeHistories = []
        with bridgedb.Storage.getDB() as db:
            for b in bridges:
                if self.cfg.COLLECT_TIMESTAMPS:
                    bh = db.getBridgeHistory(b.hex_key)
                    if bh: bridgeHistories.append(bh)
                    bridgeHistories.sort(lambda x,y: cmp(x.weightedFractionalUptime,
                                     y.weightedFractionalUptime))

            try:
                f = open(filename, 'w')
                if self.cfg.COLLECT_TIMESTAMPS:
                    for bh in bridgeHistories:
                        days = bh.tosa / long(60*60*24)
                        line = "%s:%s\t(%d days at this address)" %  \
                               (bh.ip, bh.port, days)
                        if str(bh.fingerprint) in blocklist.keys():
                            line = line + "\t(Might be blocked): (%s)" % \
                                   ",".join(blocklist[bh.fingerprint])
                        f.write(line + '\n')
                else:
                    for bridge in bridges:
                        line = "%s:%d %s" \
                               % (bridge.address, bridge.or_port, bridge.hex_key)
                        f.write(line + '\n')
                f.close()
            except IOError:
                print "I/O error: %s" % filename

    def dumpBridges(self):
        """Dump all known file distributors to files, sorted by distributor."""
        logging.info("Dumping all distributors to file.")
        with bridgedb.Storage.getDB() as db:
            allBridges = db.getAllBridges()
        bridgeDict = {}
        # Sort returned bridges by distributor
        for bridge in allBridges:
            dist = str(bridge.distributor)
            if dist in bridgeDict.keys():
                bridgeDict[dist].append(bridge)
            else:
                bridgeDict[dist] = [bridge]

        # Now dump to file(s)
        for k in bridgeDict.keys():
            dist = k
            if (dist.startswith(self.distributor_prefix)):
                # Subtract the pseudo distributor prefix
                dist = dist.replace(self.distributor_prefix, "")
            # Be safe. Replace all '/' in distributor names
            dist = dist.replace("/", "_")
            filename = dist + "-" + time.strftime("%Y-%m-%d") + ".brdgs"
            self.dumpBridgesToFile(filename, bridgeDict[k])



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2014-2015, Isis Lovecruft
#             (c) 2007-2015, The Tor Project, Inc.
#             (c) 2007-2015, all entities within the AUTHORS file
# :license: see LICENSE for licensing information

"""Unittests for the :mod:`bridgedb.parse.headers` module."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from twisted.trial import unittest

from bridgedb.parse import headers


class ParseAcceptLanguageTests(unittest.TestCase):
    """Unittests for :func:`bridgedb.parse.headers.parseAcceptLanguage`."""

    def test_noHeaders(self):
        """No header should return an empty list."""
        header = None
        langs = headers.parseAcceptLanguage(header)
        self.assertIsInstance(langs, list)
        self.assertEqual(len(langs), 0)

    def test_defaultTBBHeader(self):
        """The header 'en-us,en;q=0.5' should return ['en_us', 'en']."""
        header = 'en-us,en;q=0.5'
        langs = headers.parseAcceptLanguage(header)
        self.assertIsInstance(langs, list)
        self.assertEqual(len(langs), 2)
        self.assertEqual(langs[0], 'en_us')
        self.assertEqual(langs[1], 'en')

    def test_addNonLocalizedVariant(self):
        """The header 'en-us,en-gb;q=0.5' should return
        ['en_us', 'en', 'en_gb'].
        """
        header = 'en-us,en-gb;q=0.5'
        langs = headers.parseAcceptLanguage(header)
        self.assertIsInstance(langs, list)
        self.assertEqual(len(langs), 3)
        self.assertEqual(langs[0], 'en_us')
        self.assertEqual(langs[1], 'en')
        self.assertEqual(langs[2], 'en_gb')



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2007-2015, all entities within the AUTHORS file
#             (c) 2013-2015, Isis Lovecruft
# :license: see LICENSE for licensing information

"""Unittests for the :mod:`bridgedb.parse.nickname` module."""

from twisted.trial import unittest

from bridgedb.parse.nickname import isValidRouterNickname


class IsValidRouterNicknameTests(unittest.TestCase):
    """Unittests for :func:`bridgedb.parse.nickname.isValidRouterNickname`."""

    def test_parse_nickname_isValidRouterNickname_valid(self):
        """isValidRouterNickname() should return True for a valid nickname."""
        name = 'Unmentionable'
        self.assertTrue(isValidRouterNickname(name))

    def test_parse_nickname_isValidRouterNickname_valid_1(self):
        """isValidRouterNickname() should return True for a valid nickname."""
        name = 'maketotaldestroy'
        self.assertTrue(isValidRouterNickname(name))

    def test_parse_nickname_isValidRouterNickname_invalid_symbols(self):
        """isValidRouterNickname() should return False for an invalid nickname
        (with symbols in it).
        """
        name = 'what_the_bl#@p?!'
        self.assertFalse(isValidRouterNickname(name))

    def test_parse_nickname_isValidRouterNickname_invalid_too_long(self):
        """isValidRouterNickname() should return False for an invalid nickname
        (too long).
        """
        name = 'ThisIsReallyMoreOfANovellaRatherThanAnOnionRouterNickname'
        self.assertFalse(isValidRouterNickname(name))

    def test_parse_nickname_isValidRouterNickname_invalid_too_short(self):
        """isValidRouterNickname() should return False for an invalid nickname
        (empty string).
        """
        name = ''
        self.assertFalse(isValidRouterNickname(name))

    def test_parse_nickname_isValidRouterNickname_invalid_None(self):
        """isValidRouterNickname(None) should return False."""
        name = None
        self.assertFalse(isValidRouterNickname(name))

    def test_parse_nickname_isValidRouterNickname_invalid_spaces(self):
        """isValidRouterNickname() should return False for an invalid nickname
        (contains spaces).
        """
        name = 'As you wish'
        self.assertFalse(isValidRouterNickname(name))



# -*- coding: utf-8; test-case-name: bridgedb.test.test_email_autoresponder -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Nick Mathewson <nickm@torproject.org>
#           Isis Lovecruft <isis@torproject.org> 0xA3ADB67A2CDB8B35
#           Matthew Finkel <sysrqb@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2013-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

"""
.. py:module:: bridgedb.email.autoresponder
    :synopsis: Functionality for autoresponding to incoming emails.

bridgedb.email.autoresponder
============================

Functionality for autoresponding to incoming emails.

.. inheritance-diagram:: EmailResponse SMTPAutoresponder
    :parts: 1

::

  bridgedb.email.autoresponder
   | |_ createResponseBody - Parse lines from an incoming email and determine
   | |                       how to respond.
   | |_ generateResponse - Create an email response.
   |
   |_ EmailResponse - Holds information for generating a response to a request.
   |_ SMTPAutoresponder - An SMTP autoresponder for incoming mail.
..
"""

from __future__ import unicode_literals
from __future__ import print_function

import io
import logging
import time

from twisted.internet import defer
from twisted.internet import reactor
from twisted.mail import smtp
from twisted.python import failure

from bridgedb import safelog
from bridgedb.crypto import NEW_BUFFER_INTERFACE
from bridgedb.email import dkim
from bridgedb.email import request
from bridgedb.email import templates
from bridgedb.email.distributor import EmailRequestedHelp
from bridgedb.email.distributor import EmailRequestedKey
from bridgedb.email.distributor import TooSoonEmail
from bridgedb.email.distributor import IgnoreEmail
from bridgedb.parse import addr
from bridgedb.parse.addr import canonicalizeEmailDomain
from bridgedb.util import levenshteinDistance
from bridgedb import translations


def createResponseBody(lines, context, client, lang='en'):
    """Parse the **lines** from an incoming email request and determine how to
    respond.

    :param list lines: The list of lines from the original request sent by the
        client.
    :type context: :class:`bridgedb.email.server.MailServerContext`
    :param context: The context which contains settings for the email server.
    :type client: :api:`twisted.mail.smtp.Address`
    :param client: The client's email address which should be in the
        ``'To:'`` header of the response email.
    :param str lang: The 2-5 character locale code to use for translating the
        email. This is obtained from a client sending a email to a valid plus
        address which includes the translation desired, i.e. by sending an
        email to `bridges+fa@torproject.org
        <mailto:bridges+fa@torproject.org>`__, the client should receive a
        response in Farsi.
    :rtype: str
    :returns: ``None`` if we shouldn't respond to the client (i.e., if they
        have already received a rate-limiting warning email). Otherwise,
        returns a string containing the (optionally translated) body for the
        email response which we should send out.
    """
    translator = translations.installTranslations(lang)
    bridges = None
    try:
        bridgeRequest = request.determineBridgeRequestOptions(lines)
        bridgeRequest.client = str(client)

        # The request was invalid, respond with a help email which explains
        # valid email commands:
        if not bridgeRequest.isValid():
            raise EmailRequestedHelp("Email request from '%s' was invalid."
                                     % str(client))

        # Otherwise they must have requested bridges:
        interval = context.schedule.intervalStart(time.time())
        bridges = context.distributor.getBridges(bridgeRequest, interval)
    except EmailRequestedHelp as error:
        logging.info(error)
        return templates.buildWelcomeText(translator, client)
    except EmailRequestedKey as error:
        logging.info(error)
        return templates.buildKeyMessage(translator, client)
    except TooSoonEmail as error:
        logging.info("Got a mail too frequently: %s." % error)
        return templates.buildSpamWarning(translator, client)
    except (IgnoreEmail, addr.BadEmail) as error:
        logging.info(error)
        # Don't generate a response if their email address is unparsable or
        # invalid, or if we've already warned them about rate-limiting:
        return None
    else:
        answer = "(no bridges currently available)\r\n"
        if bridges:
            transport = bridgeRequest.justOnePTType()
            answer = "".join("  %s\r\n" % b.getBridgeLine(
                bridgeRequest, context.includeFingerprints) for b in bridges)
        return templates.buildAnswerMessage(translator, client, answer)

def generateResponse(fromAddress, client, body, subject=None,
                     messageID=None, gpgSignFunc=None):
    """Create an :class:`EmailResponse`, which acts like an
    :class:`io.StringIO` instance, by creating and writing all headers and the
    email body into the file-like :attr:`EmailResponse.mailfile`.

    :param str fromAddress: The :rfc:`2821` email address which should be in
        the ``'From:'`` header.
    :type client: :api:`twisted.mail.smtp.Address`
    :param client: The client's email address which should be in the
        ``'To:'`` header of the response email.
    :param str subject: The string to write to the ``'Subject:'`` header.
    :param str body: The body of the email. If a **gpgSignFunc** is also
        given, then :meth:`EmailResponse.writeBody` will generate and include
        an ascii-armored OpenPGP signature in the **body**.
    :type messageID: ``None`` or :any:`str`
    :param messageID: The :rfc:`2822` specifier for the ``'Message-ID:'``
        header, if including one is desirable.
    :type gpgSignFunc: callable
    :param gpgSignFunc: If given, this should be a callable function for
        signing messages.  See :func:`bridgedb.crypto.initializeGnuPG` for
        obtaining a pre-configured **gpgSignFunc**.
    :returns: An :class:`EmailResponse` which contains the entire email. To
        obtain the contents of the email, including all headers, simply use
        :meth:`EmailResponse.readContents`.
    """
    response = EmailResponse(gpgSignFunc)
    response.to = client
    response.writeHeaders(fromAddress.encode('utf-8'), str(client), subject,
                          inReplyTo=messageID)
    response.writeBody(body.encode('utf-8'))

    # Only log the email text (including all headers) if SAFE_LOGGING is
    # disabled:
    if not safelog.safe_logging:
        contents = response.readContents()
        logging.debug("Email contents:\n%s" % str(contents))
    else:
        logging.debug("Email text for %r created." % str(client))

    response.rewind()
    return response


class EmailResponse(object):
    """Holds information for generating a response email for a request.

    .. todo:: At some point, we may want to change this class to optionally
        handle creating Multipart MIME encoding messages, so that we can
        include attachments. (This would be useful for attaching our GnuPG
        keyfile, for example, rather than simply pasting it into the body of
        the email.)


    :vartype _buff: :any:`unicode` or :any:`buffer`
    :var _buff: Used internally to write lines for the response email into the
        ``_mailfile``. The reason why both of these attributes have two
        possible types is for the same Python-buggy reasons which require
        :data:`~bridgedb.crypto.NEW_BUFFER_INTERFACE`.
    :vartype mailfile: :class:`io.StringIO` or :class:`io.BytesIO`
    :var mailfile: An in-memory file-like object for storing the formatted
        headers and body of the response email.
    :var str delimiter: Delimiter between lines written to the
        :data:`mailfile`.
    :var bool closed: ``True`` if :meth:`close` has been called.
    :vartype to: :api:`twisted.mail.smtp.Address`
    :var to: The client's email address, to which this response should be sent.
    """
    _buff = buffer if NEW_BUFFER_INTERFACE else unicode
    mailfile = io.BytesIO if NEW_BUFFER_INTERFACE else io.StringIO

    def __init__(self, gpgSignFunc=None):
        """Create a response to an email we have recieved.

        This class deals with correctly formatting text for the response email
        headers and the response body into an instance of :data:`mailfile`.

        :type gpgSignFunc: callable
        :param gpgSignFunc: If given, this should be a callable function for
            signing messages.  See :func:`bridgedb.crypto.initializeGnuPG` for
            obtaining a pre-configured **gpgSignFunc**.
        """
        self.gpgSign = gpgSignFunc
        self.mailfile = self.mailfile()
        self.delimiter = '\n'
        self.closed = False
        self.to = None

    def close(self):
        """Close our :data:`mailfile` and set :data:`closed` to ``True``."""
        logging.debug("Closing %s.mailfile..." % (self.__class__.__name__))
        self.mailfile.close()
        self.closed = True

    def read(self, size=None):
        """Read, at most, **size** bytes from our :data:`mailfile`.

        .. note:: This method is required by Twisted's SMTP system.

        :param int size: The number of bytes to read. Defaults to ``None``,
            which reads until EOF.
        :rtype: str
        :returns: The bytes read from the :data:`mailfile`.
        """
        contents = ''
        logging.debug("Reading%s from %s.mailfile..."
                      % ((' {0} bytes'.format(size) if size else ''),
                         self.__class__.__name__))
        try:
            if size is not None:
                contents = self.mailfile.read(int(size)).encode('utf-8')
            else:
                contents = self.mailfile.read().encode('utf-8')
        except Exception as error:  # pragma: no cover
            logging.exception(error)

        return contents

    def readContents(self):
        """Read the all the contents written thus far to the :data:`mailfile`,
        and then :meth:`seek` to return to the original pointer position we
        were at before this method was called.

        :rtype: str
        :returns: The entire contents of the :data:`mailfile`.
        """
        pointer = self.mailfile.tell()
        self.mailfile.seek(0)
        contents = self.mailfile.read()
        self.mailfile.seek(pointer)
        return contents

    def rewind(self):
        """Rewind to the very beginning of the :data:`mailfile`."""
        logging.debug("Rewinding %s.mailfile..." % self.__class__.__name__)
        self.mailfile.seek(0)

    def write(self, line):
        """Write the **line** to the :data:`mailfile`.

        Any **line** written to me will have :data:`delimiter` appended to it
        beforehand.

        :param str line: Something to append into the :data:`mailfile`.
        """
        if line.find('\r\n') != -1:
            # If **line** contains newlines, send it to :meth:`writelines` to
            # break it up so that we can replace them:
            logging.debug("Found newlines in %r. Calling writelines()." % line)
            self.writelines(line)
        else:
            line += self.delimiter
            self.mailfile.write(self._buff(line.encode('utf8')))
            self.mailfile.flush()

    def writelines(self, lines):
        """Calls :meth:`write` for each line in **lines**.

        Line endings of ``'\\r\\n'`` will be replaced with :data:`delimiter`
        (i.e. ``'\\n'``). See :api:`twisted.mail.smtp.SMTPClient.getMailData`
        for the reason.

        :type lines: :any:`basestring` or :any:`list`
        :param lines: The lines to write to the :attr:`mailfile`.
        """
        if isinstance(lines, basestring):
            lines = lines.replace('\r\n', '\n')
            for ln in lines.split('\n'):
                self.write(ln)
        elif isinstance(lines, (list, tuple,)):
            for ln in lines:
                self.write(ln)

    def writeHeaders(self, fromAddress, toAddress, subject=None,
                     inReplyTo=None, includeMessageID=True,
                     contentType='text/plain; charset="utf-8"', **kwargs):
        """Write all headers into the response email.

        :param str fromAddress: The email address for the ``'From:'`` header.
        :param str toAddress: The email address for the ``'To:'`` header.
        :type subject: ``None`` or :any:`str`
        :param subject: The ``'Subject:'`` header.
        :type inReplyTo: ``None`` or :any:`str`
        :param inReplyTo: If set, an ``'In-Reply-To:'`` header will be
            generated. This should be set to the ``'Message-ID:'`` header from
            the client's original request email.
        :param bool includeMessageID: If ``True``, generate and include a
            ``'Message-ID:'`` header for the response.
        :param str contentType: The ``'Content-Type:'`` header.
        :kwargs: If given, the key will become the name of the header, and the
            value will become the contents of that header.
        """
        self.write("From: %s" % fromAddress)
        self.write("To: %s" % toAddress)
        if includeMessageID:
            self.write("Message-ID: %s" % smtp.messageid().encode('utf-8'))
        if inReplyTo:
            self.write("In-Reply-To: %s" % inReplyTo.encode('utf-8'))
        self.write("Content-Type: %s" % contentType.encode('utf-8'))
        self.write("Date: %s" % smtp.rfc822date().encode('utf-8'))

        if not subject:
            subject = '[no subject]'
        if not subject.lower().startswith('re'):
            subject = "Re: " + subject
        self.write("Subject: %s" % subject.encode('utf-8'))

        if kwargs:
            for headerName, headerValue in kwargs.items():
                headerName = headerName.capitalize()
                headerName = headerName.replace(' ', '-')
                headerName = headerName.replace('_', '-')
                header = "%s: %s" % (headerName, headerValue)
                self.write(header.encode('utf-8'))

        # The first blank line designates that the headers have ended:
        self.write(self.delimiter)

    def writeBody(self, body):
        """Write the response body into the :attr:`mailfile`.

        If :attr:`EmailResponse.gpgSignFunc` is set, and signing is configured,
        the **body** will be automatically signed before writing its contents
        into the :attr:`mailfile`.

        :param str body: The body of the response email.
        """
        logging.info("Writing email body...")
        if self.gpgSign:
            logging.info("Attempting to sign email...")
            sig = self.gpgSign(body)
            if sig:
                body = sig
        self.writelines(body)


class SMTPAutoresponder(smtp.SMTPClient):
    """An :api:`twisted.mail.smtp.SMTPClient` for responding to incoming mail.

    The main worker in this class is the :meth:`reply` method, which functions
    to dissect an incoming email from an incoming :class:`SMTPMessage` and
    create a :class:`EmailResponse` email message in reply to it, and then,
    finally, send it out.

    :vartype log: :api:`twisted.python.util.LineLog`
    :ivar log: A cache of debug log messages.
    :vartype debug: bool
    :ivar debug: If ``True``, enable logging (accessible via :attr:`log`).
    :ivar str identity: Our FQDN which will be sent during client ``HELO``.
    :vartype incoming: :api:`Message <twisted.mail.smtp.rfc822.Message>`
    :ivar incoming: An incoming message, i.e. as returned from
        :meth:`SMTPMessage.getIncomingMessage`.

    :vartype deferred: :api:`twisted.internet.defer.Deferred`

    :ivar deferred: A :api:`Deferred <twisted.internet.defer.Deferred>` with
       registered callbacks, :meth:`sentMail` and :meth:`sendError`, which
       will be given to the reactor in order to process the sending of the
       outgoing response email.
    """
    debug = True
    identity = smtp.DNSNAME

    def __init__(self):
        """Handle responding (or not) to an incoming email."""
        smtp.SMTPClient.__init__(self, self.identity)
        self.incoming = None
        self.deferred = defer.Deferred()
        self.deferred.addCallback(self.sentMail)
        self.deferred.addErrback(self.sendError)

    def getMailData(self):
        """Gather all the data for building the response to the client.

        This method must return a file-like object containing the data of the
        message to be sent. Lines in the file should be delimited by ``\\n``.

        :rtype: ``None`` or :class:`EmailResponse`
        :returns: An ``EmailResponse``, if we have a response to send in reply
            to the incoming email, otherwise, returns ``None``.
        """
        clients = self.getMailTo()
        if not clients: return
        client = clients[0]  # There should have been only one anyway

        # Log the email address that this message came from if SAFELOGGING is
        # not enabled:
        if not safelog.safe_logging:
            logging.debug("Incoming email was from %s ..." % client)

        if not self.runChecks(client): return

        recipient = self.getMailFrom()
        # Look up the locale part in the 'To:' address, if there is one, and
        # get the appropriate Translation object:
        lang = translations.getLocaleFromPlusAddr(recipient)
        logging.info("Client requested email translation: %s" % lang)

        body = createResponseBody(self.incoming.lines,
                                  self.incoming.context,
                                  client, lang)
        if not body: return  # The client was already warned.

        messageID = self.incoming.message.getheader("Message-ID", None)
        subject = self.incoming.message.getheader("Subject", None)
        response = generateResponse(recipient, client,
                                    body, subject, messageID,
                                    self.incoming.context.gpgSignFunc)
        return response

    def getMailTo(self):
        """Attempt to get the client's email address from an incoming email.

        :rtype: list
        :returns: A list containing the client's
            :func:`normalized <bridgedb.parse.addr.normalizeEmail>` email
            :api:`Address <twisted.mail.smtp.Address>`, if it originated from
            a domain that we accept and the address was well-formed. Otherwise,
            returns ``None``. Even though we're likely to respond to only one
            client at a time, the return value of this method must be a list
            in order to hook into the rest of
            :api:`twisted.mail.smtp.SMTPClient` correctly.
        """
        clients = []
        addrHeader = None
        try: fromAddr = self.incoming.message.getaddr("From")[1]
        except (IndexError, TypeError, AttributeError): pass
        else: addrHeader = fromAddr

        if not addrHeader:
            logging.warn("No From header on incoming mail.")
            try: senderHeader = self.incoming.message.getaddr("Sender")[1]
            except (IndexError, TypeError, AttributeError): pass
            else: addrHeader = senderHeader
        if not addrHeader:
            logging.warn("No Sender header on incoming mail.")
            return clients

        client = None
        try:
            if addrHeader in self.incoming.context.whitelist.keys():
                logging.debug("Email address was whitelisted: %s."
                              % addrHeader)
                client = smtp.Address(addrHeader)
            else:
                normalized = addr.normalizeEmail(
                    addrHeader,
                    self.incoming.context.domainMap,
                    self.incoming.context.domainRules)
                client = smtp.Address(normalized)
        except (addr.UnsupportedDomain) as error:
            logging.warn(error)
        except (addr.BadEmail, smtp.AddressError) as error:
            logging.warn(error)

        if client:
            clients.append(client)

        return clients

    def getMailFrom(self):
        """Find our address in the recipients list of the **incoming** message.

        :rtype: str
        :return: Our address from the recipients list. If we can't find it
            return our default ``EMAIL_FROM_ADDRESS`` from the config file.
        """
        logging.debug("Searching for our email address in 'To:' header...")

        ours = None

        try:
            ourAddress = smtp.Address(self.incoming.context.fromAddr)
            allRecipients = self.incoming.message.getaddrlist("To")

            for _, addr in allRecipients:
                recipient = smtp.Address(addr)
                if not ourAddress.domain in recipient.domain:
                    logging.debug(("Not our domain (%s) or subdomain, skipping"
                                   " email address: %s")
                                  % (ourAddress.domain, str(recipient)))
                    continue
                # The recipient's username should at least start with ours,
                # but it still might be a '+' address.
                if not recipient.local.startswith(ourAddress.local):
                    logging.debug(("Username doesn't begin with ours, skipping"
                                   " email address: %s") % str(recipient))
                    continue
                # Only check the username before the first '+':
                beforePlus = recipient.local.split('+', 1)[0]
                if beforePlus == ourAddress.local:
                    ours = str(recipient)
            if not ours:
                raise addr.BadEmail(allRecipients)

        except Exception as error:
            logging.error(("Couldn't find our email address in incoming email "
                           "headers: %r" % error))
            # Just return the email address that we're configured to use:
            ours = self.incoming.context.fromAddr

        logging.debug("Found our email address: %s." % ours)
        return ours

    def sentMail(self, success):
        """Callback for a :api:`twisted.mail.smtp.SMTPSenderFactory`,
        called when an attempt to send an email is completed.

        If some addresses were accepted, code and resp are the response
        to the DATA command. If no addresses were accepted, code is -1
        and resp is an informative message.

        :param int code: The code returned by the SMTP Server.
        :param str resp: The string response returned from the SMTP Server.
        :param int numOK: The number of addresses accepted by the remote host.
        :param list addresses: A list of tuples (address, code, resp) listing
            the response to each ``RCPT TO`` command.
        :param log: The SMTP session log. (We don't use this, but it is sent
            by :api:`twisted.mail.smtp.SMTPSenderFactory` nonetheless.)
        """
        numOk, addresses = success

        for (address, code, resp) in addresses:
            logging.info("Sent reply to %s" % address)
            logging.debug("SMTP server response: %d %s" % (code, resp))

        if self.debug:
            for line in self.log.log:
                if line:
                    logging.debug(line)

    def sendError(self, fail):
        """Errback for a :api:`twisted.mail.smtp.SMTPSenderFactory`.

        :type fail: :api:`twisted.python.failure.Failure` or
            :api:`twisted.mail.smtp.SMTPClientError`
        :param fail: An exception which occurred during the transaction to
            send the outgoing email.
        """
        logging.debug("called with %r" % fail)

        if isinstance(fail, failure.Failure):
            error = fail.getTraceback() or "Unknown"
        elif isinstance(fail, Exception):
            error = fail
        logging.error(error)

        try:
            # This handles QUIT commands, disconnecting, and closing the
            # transport:
            smtp.SMTPClient.sendError(self, fail)
        # We might not have `transport` and `protocol` attributes, depending
        # on when and where the error occurred, so just catch and log it:
        except Exception as error:
            logging.error(error)

    def reply(self):
        """Reply to an incoming email. Maybe.

        If nothing is returned from either :func:`createResponseBody` or
        :func:`generateResponse`, then the incoming email will not be
        responded to at all. This can happen for several reasons, for example:
        if the DKIM signature was invalid or missing, or if the incoming email
        came from an unacceptable domain, or if there have been too many
        emails from this client in the allotted time period.

        :rtype: :api:`twisted.internet.defer.Deferred`
        :returns: A ``Deferred`` which will callback when the response has
            been successfully sent, or errback if an error occurred while
            sending the email.
        """
        logging.info("Got an email; deciding whether to reply.")

        response = self.getMailData()
        if not response:
            return self.deferred

        return self.send(response)

    def runChecks(self, client):
        """Run checks on the incoming message, and only reply if they pass.

        1. Check if the client's address is whitelisted.

        2. If it's not whitelisted, check that the domain names, taken from
        the SMTP ``MAIL FROM:`` command and the email ``'From:'`` header, can
        be :func:`canonicalized <addr.canonicalizeEmailDomain>`.

        3. Check that those canonical domains match.

        4. If the incoming message is from a domain which supports DKIM
        signing, then run :func:`bridgedb.email.dkim.checkDKIM` as well.

        .. note:: Calling this method sets the
            :attr:`incoming.canonicalFromEmail` and
            :attr:`incoming.canonicalDomainRules` attributes of the
            :attr:`incoming` message.

        :type client: :api:`twisted.mail.smtp.Address`
        :param client: The client's email address, extracted from the
            ``'From:'`` header from the incoming email.
        :rtype: bool
        :returns: ``False`` if the checks didn't pass, ``True`` otherwise.
        """
        # If the SMTP ``RCPT TO:`` domain name couldn't be canonicalized, then
        # we *should* have bailed at the SMTP layer, but we'll reject this
        # email again nonetheless:
        if not self.incoming.canonicalFromSMTP:
            logging.warn(("SMTP 'MAIL FROM' wasn't from a canonical domain "
                          "for email from %s") % str(client))
            return False

        # Allow whitelisted addresses through the canonicalization check:
        if str(client) in self.incoming.context.whitelist.keys():
            self.incoming.canonicalFromEmail = client.domain
            logging.info("'From:' header contained whitelisted address: %s"
                         % str(client))
        # Straight up reject addresses in the EMAIL_BLACKLIST config option:
        elif str(client) in self.incoming.context.blacklist:
            logging.info("'From:' header contained blacklisted address: %s")
            return False
        else:
            logging.debug("Canonicalizing client email domain...")
            try:
                # The client's address was already checked to see if it came
                # from a supported domain and is a valid email address in
                # :meth:`getMailTo`, so we should just be able to re-extract
                # the canonical domain safely here:
                self.incoming.canonicalFromEmail = canonicalizeEmailDomain(
                    client.domain, self.incoming.canon)
                logging.debug("Canonical email domain: %s"
                              % self.incoming.canonicalFromEmail)
            except addr.UnsupportedDomain as error:
                logging.info("Domain couldn't be canonicalized: %s"
                             % safelog.logSafely(client.domain))
                return False

        # The canonical domains from the SMTP ``MAIL FROM:`` and the email
        # ``From:`` header should match:
        if self.incoming.canonicalFromSMTP != self.incoming.canonicalFromEmail:
            logging.error("SMTP/Email canonical domain mismatch!")
            logging.debug("Canonical domain mismatch: %s != %s"
                          % (self.incoming.canonicalFromSMTP,
                             self.incoming.canonicalFromEmail))
            #return False

        self.incoming.domainRules = self.incoming.context.domainRules.get(
            self.incoming.canonicalFromEmail, list())

        # If the domain's ``domainRules`` say to check DKIM verification
        # results, and those results look bad, reject this email:
        if not dkim.checkDKIM(self.incoming.message, self.incoming.domainRules):
            return False

        # If fuzzy matching is enabled via the EMAIL_FUZZY_MATCH setting, then
        # calculate the Levenshtein String Distance (see
        # :func:`~bridgedb.util.levenshteinDistance`):
        if self.incoming.context.fuzzyMatch != 0:
            for blacklistedAddress in self.incoming.context.blacklist:
                distance = levenshteinDistance(str(client), blacklistedAddress)
                if distance <= self.incoming.context.fuzzyMatch:
                    logging.info("Fuzzy-matched %s to blacklisted address %s!"
                                 % (self.incoming.canonicalFromEmail,
                                    blacklistedAddress))
                    return False

        return True

    def send(self, response, retries=0, timeout=30, reaktor=reactor):
        """Send our **response** in reply to :data:`incoming`.

        :type client: :api:`twisted.mail.smtp.Address`
        :param client: The email address of the client.
        :param response: A :class:`EmailResponse`.
        :param int retries: Try resending this many times. (default: ``0``)
        :param int timeout: Timeout after this many seconds. (default: ``30``)
        :rtype: :api:`Deferred <twisted.internet.defer.Deferred>`
        :returns: Our :data:`deferred`.
        """
        logging.info("Sending reply to %s ..." % str(response.to))

        factory = smtp.SMTPSenderFactory(self.incoming.context.smtpFromAddr,
                                         str(response.to),
                                         response,
                                         self.deferred,
                                         retries=retries,
                                         timeout=timeout)
        factory.domain = smtp.DNSNAME
        reaktor.connectTCP(self.incoming.context.smtpServerIP,
                           self.incoming.context.smtpServerPort,
                           factory)
        return self.deferred



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_interfaces ; -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2014-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________


"""All available `Zope`_ interfaces in BridgeDB.

.. _Zope: http://docs.zope.org/zope.interface/index.html
"""

from zope.interface import Interface
from zope.interface import Attribute
from zope.interface import implementer


class IName(Interface):
    """An interface specification for a named object."""

    name = Attribute("A string which identifies this object.")


@implementer(IName)
class Named(object):
    """A named object."""

    #: The character(s) used to join child :class:`Named` object's names with
    #: our name.
    separator = ' '

    def __init__(self):
        self._name = str()

    @property
    def name(self):
        """Get the name of this object.

        :rtype: str
        :returns: A string which identifies this object.
        """
        return self._name

    @name.setter
    def name(self, name):
        """Set a **name** for identifying this object.

        This is used to identify the object in log messages; the **name**
        doesn't necessarily need to be unique. Other :class:`Named` objects
        which are properties of a :class:`Named` object may inherit their
        parents' **name**s.

        >>> from bridgedb.distribute import Named
        >>> named = Named()
        >>> named.name = 'Excellent Super-Awesome Thing'
        >>> named.name
        'Excellent Super-Awesome Thing'

        :param str name: A name for this object.
        """
        self._name = name

        for attr in self.__dict__.values():
            if IName.providedBy(attr):
                attr.name = self.separator.join([name, attr.name])



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_persistent -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013-2015 Isis Lovecruft
#             (c) 2007-2015, The Tor Project, Inc.
#             (c) 2007-2015, all entities within the AUTHORS file
# :license: 3-clause BSD, see included LICENSE for information

"""Module for functionality to persistently store state."""

import copy
import logging
import os.path

try:
    import cPickle as pickle
except (ImportError, NameError):  # pragma: no cover
    import pickle

from twisted.python.reflect import safe_repr
from twisted.spread import jelly

from bridgedb import Bridges
from bridgedb import filters
from bridgedb.email import distributor as emailDistributor
from bridgedb.https import distributor as httpsDistributor
from bridgedb.configure import Conf
#from bridgedb.proxy import ProxySet

_state = None

#: Types and classes which are allowed to be jellied:
_security = jelly.SecurityOptions()
#_security.allowInstancesOf(ProxySet)
_security.allowModules(filters, Bridges, emailDistributor, httpsDistributor)


class MissingState(Exception):
    """Raised when the file or class storing global state is missing."""


def _getState():
    """Retrieve the global state instance.

    :rtype: :class:`~bridgedb.persistent.State`
    :returns: An unpickled de-sexp'ed state object, which may contain just
        about anything, but should contain things like options, loaded config
        settings, etc.
    """
    return _state

def _setState(state):
    """Set the global state.

    :type state: :class:`~bridgedb.persistent.State`
    :param state: The state instance to save.
    """
    global _state
    _state = state

def load(stateCls=None):
    """Given a :class:`State`, try to unpickle it's ``statefile``.

    :param string stateCls: An instance of :class:`~bridgedb.persistent.State`. If
        not given, try loading from ``_statefile`` if that file exists.
    :rtype: None or :class:`State`
    """
    statefile = None
    if stateCls and isinstance(stateCls, State):
        cls = stateCls
    else:
        cls = _getState()

    if not cls:
        raise MissingState("Could not find a state instance to load.")
    else:
        loaded = cls.load()
        return loaded


class State(jelly.Jellyable):
    """Pickled, jellied storage container for persistent state."""

    def __init__(self, config=None, **kwargs):
        """Create a persistent state storage mechanism.

        Serialisation of certain classes in BridgeDB doesn't work. Classes and
        modules which are known to be unjelliable/unpicklable so far are:

          - bridgedb.Dist
          - bridgedb.Bridges, and all "splitter" and "ring" classes contained
            within

        :property statefile: The filename to retrieve a pickled, jellied
            :class:`~bridgedb.persistent.State` instance from. (default:
            :attr:`bridgedb.persistent.State._statefile`)
        """
        self._statefile = os.path.abspath(str(__package__) + '.state')
        self.proxyList = None
        self.config = None
        self.key = None

        if 'STATEFILE' in kwargs:
            self.statefile = kwargs['STATEFILE']

        for key, value in kwargs.items():
            self.__dict__[key] = value

        if config is not None:
            for key, value in config.__dict__.items():
                self.__dict__[key] = value

        _setState(self)

    def _get_statefile(self):
        """Retrieve the filename of the global statefile.

        :rtype: string
        :returns: The filename of the statefile.
        """
        return self._statefile

    def _set_statefile(self, filename):
        """Set the global statefile.

        :param string statefile: The filename of the statefile.
        """
        filename = os.path.abspath(os.path.expanduser(filename))
        logging.debug("Setting statefile to '%s'" % filename)
        self._statefile = filename

        # Create the parent directory if it doesn't exist:
        dirname = os.path.dirname(filename)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)

        # Create the statefile if it doesn't exist:
        if not os.path.exists(filename):
            open(filename, 'w').close()

    def _del_statefile(self):
        """Delete the file containing previously saved state."""
        try:
            with open(self._statefile, 'w') as fh:
                fh.close()
            os.unlink(self._statefile)
            self._statefile = None
        except (IOError, OSError) as error:  # pragma: no cover
            logging.error("There was an error deleting the statefile: '%s'"
                          % self._statefile)

    statefile = property(_get_statefile, _set_statefile, _del_statefile,
                         """Filename property of a persisent.State.""")

    def load(self, statefile=None):
        """Load a previously saved statefile.

        :raises MissingState: If there was any error loading the **statefile**.
        :rtype: :class:`State` or None
        :returns: The state, loaded from :attr:`State.STATEFILE`, or None if
            an error occurred.
        """
        if not statefile:
            if not self.statefile:
                raise MissingState("Could not find a state file to load.")
            statefile = self.statefile
        logging.debug("Retrieving state from: \t'%s'" % statefile)

        quo= fh = None
        err = ''

        try:
            if isinstance(statefile, basestring):
                fh = open(statefile, 'r')
            elif not statefile.closed:
                fh = statefile
        except (IOError, OSError) as error:  # pragma: no cover
            err += "There was an error reading statefile "
            err += "'{0}':\n{1}".format(statefile, error)
        except (AttributeError, TypeError) as error:
            err += "Failed statefile.open() and statefile.closed:"
            err += "\n\t{0}\nstatefile type = '{1}'".format(
                error.message, type(statefile))
        else:
            try:
                status = pickle.load(fh)
            except EOFError:
                err += "The statefile %s was empty." % fh.name
            else:
                quo = jelly.unjelly(status)
                if fh is not None:
                    fh.close()
                if quo:
                    return quo

        if err:
            raise MissingState(err)

    def save(self, statefile=None):
        """Save state as a pickled jelly to a file on disk."""
        if not statefile:
            if not self._statefile:
                raise MissingState("Could not find a state file to load.")
            statefile = self._statefile
        logging.debug("Saving state to: \t'%s'" % statefile)

        fh = None
        try:
            fh = open(statefile, 'w')
        except (IOError, OSError) as error:  # pragma: no cover
            logging.warn("Error writing state file to '%s': %s"
                         % (statefile, error))
        else:
            try:
                pickle.dump(jelly.jelly(self), fh)
            except AttributeError as error:
                logging.debug("Tried jellying an unjelliable object: %s"
                              % error.message)

        if fh is not None:
            fh.flush()
            fh.close()

    def useChangedSettings(self, config):
        """Take a new config, compare it to the last one, and update settings.

        Given a ``config`` object created from the configuration file, compare
        it to the last :class:`~bridgedb.configure.Conf` that was stored, and apply
        any settings which were changed to be attributes of the :class:`State`
        instance.
        """
        updated = []
        new = []

        for key, value in config.__dict__.items():
            try:
                # If state.config doesn't have the same value as the new
                # config, then update the state setting.
                #
                # Be sure, when updating settings while parsing the config
                # file, to assign the new settings as attributes of the
                # :class:`bridgedb.configure.Conf` instance.
                if value != self.config.__dict__[key]:
                    setattr(self, key, value)
                    updated.append(key)
                    logging.debug("Updated %s setting: %r → %r" %
                                  (safe_repr(key),
                                   self.config.__dict__[key],
                                   safe_repr(value)))
            except (KeyError, AttributeError):
                setattr(self, key, value)
                new.append(key)
                logging.debug("New setting: %s = %r" %
                              (safe_repr(key),
                               safe_repr(value)))

        logging.info("Updated setting(s): %s" % ' '.join([x for x in updated]))
        logging.info("New setting(s): %s" % ' '.join([x for x in new]))
        logging.debug(
            "Saving newer config as `state.config` for later comparison")
        self.config = config
